# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable
from ctypes import Structure, cast, POINTER
from ctypes.wintypes import POINT

from xpra.util.env import envbool
from xpra.common import noop
from xpra.platform.win32.wndproc_events import WNDPROC_EVENT_NAMES
from xpra.platform.win32 import constants as win32con
from xpra.platform.win32.common import WNDPROC, GetWindowLongW, SetWindowLongW, GetSystemMetrics, CallWindowProcW
from xpra.log import Logger


log = Logger("win32", "window", "util")
vlog = Logger("verbose")


class MINMAXINFO(Structure):
    _fields_ = [
        ("ptReserved", POINT),
        ("ptMaxSize", POINT),
        ("ptMaxPosition", POINT),
        ("ptMinTrackSize", POINT),
        ("ptMaxTrackSize", POINT),
    ]


# loosely based on this recipe:
# http://code.activestate.com/recipes/334779-pygtk-win32-extension-empower-gtk-with-win32-windo/
# and this WM_GETMINMAXINFO ctypes code:
# https://github.com/Mozillion/SublimeSpeech/blob/master/lib/dragonfly/windows/dialog_base.py
# only hardcoded for handling WM_GETMINMAXINFO,
# but should be pretty easy to tweak if needed.

HOOK_MINMAXINFO = envbool("XPRA_WIN32_MINMAXINFO", True)
HOOK_MINMAXINFO_OVERRIDE = envbool("XPRA_WIN32_MINMAXINFO_OVERRIDE", True)


class Win32Hooks:

    def __init__(self, hwnd: int):
        self._hwnd = hwnd
        self._message_map: dict[int, Callable[[int, int, int, int], int]] = {}
        self.max_size = (0, 0)
        self.min_size = (0, 0)
        self.frame_width = 4
        self.frame_height = 4
        self.caption_height = 26
        self._oldwndproc = None
        self._newwndproc = None
        if HOOK_MINMAXINFO:
            self.add_window_event_handler(win32con.WM_GETMINMAXINFO, self.on_getminmaxinfo)
        try:
            # we only use this code for resizable windows, so use SM_C?SIZEFRAME:
            self.frame_width = GetSystemMetrics(win32con.SM_CXSIZEFRAME)
            self.frame_height = GetSystemMetrics(win32con.SM_CYSIZEFRAME)
            self.caption_height = GetSystemMetrics(win32con.SM_CYCAPTION)
        except Exception:
            log("error querying frame attributes", exc_info=True)
        log("Win32Hooks: window frame size is %sx%s", self.frame_width, self.frame_height)
        log("Win32Hooks: message_map=%s", self._message_map)

    def add_window_event_handler(self, event: int, handler: Callable[[int, int, int, int], int]) -> None:
        self._message_map[event] = handler

    def setup(self) -> None:
        assert self._oldwndproc is None
        self._newwndproc = WNDPROC(self._wndproc)
        try:
            self._oldwndproc = SetWindowLongW(self._hwnd, win32con.GWL_WNDPROC, self._newwndproc)
        except Exception:
            log.error(f"Error setting up window hook for {self._hwnd}", exc_info=True)

    def on_getminmaxinfo(self, hwnd: int, msg, wparam: int, lparam: int) -> int:
        if (self.min_size > (0, 0) or self.max_size > (0, 0)) and lparam:
            info = cast(lparam, POINTER(MINMAXINFO)).contents
            style = GetWindowLongW(hwnd, win32con.GWL_STYLE)
            dw, dh = 0, 0
            if style & win32con.WS_BORDER:
                fw, fh = self.frame_width, self.frame_height
                dw = fw * 2
                dh = self.caption_height + fh * 2
                log("on_getminmaxinfo window=%#x min_size=%s, max_size=%s, frame=%sx%s",
                    hwnd, self.min_size, self.max_size, fw, fh)
            if self.min_size > (0, 0):
                minw, minh = self.min_size  # pylint: disable=unpacking-non-sequence
                minw += dw
                minh += dh
                if not HOOK_MINMAXINFO_OVERRIDE:
                    v = info.ptMinTrackSize
                    x = int(v.x)
                    y = int(v.y)
                    if v:
                        log("on_getminmaxinfo ptMinTrackSize=%ix%i", v.x, v.y)
                        if x > 0:
                            minw = max(minw, x)
                        if y > 0:
                            minh = max(minh, y)
                point = POINT(minw, minh)
                info.ptMinSize = point
                info.ptMinTrackSize = point
                log("on_getminmaxinfo actual min_size=%ix%i", minw, minh)
            if self.max_size > (0, 0):
                maxw, maxh = self.max_size  # pylint: disable=unpacking-non-sequence
                maxw += dw
                maxh += dh
                if not HOOK_MINMAXINFO_OVERRIDE:
                    for name, v in {
                        "ptMaxSize": info.ptMaxSize,
                        "ptMaxTrackSize": info.ptMaxTrackSize,
                    }.items():
                        if v:
                            x = int(v.x)
                            y = int(v.y)
                            log("on_getminmaxinfo %s=%ix%i", name, v.x, v.y)
                            if x > 0:
                                maxw = min(maxw, x)
                            if y > 0:
                                maxh = min(maxh, y)
                point = POINT(maxw, maxh)
                info.ptMaxSize = point
                info.ptMaxTrackSize = point
                log("on_getminmaxinfo actual max_size=%ix%i", maxw, maxh)
            return 0
        log("on_getminmaxinfo%s min_size=%s, max_size=%s",
            (hwnd, msg, wparam, lparam), self.min_size, self.max_size)
        return 0

    def cleanup(self, *args) -> None:
        log("cleanup%s", args)
        self._message_map = {}
        # since we assume the window is closed, restoring the wnd proc may be redundant here:
        if not self._oldwndproc or not self._hwnd:
            return
        with log.trap_error("Error: window hooks cleanup failure"):
            SetWindowLongW(self._hwnd, win32con.GWL_WNDPROC, self._oldwndproc)
            self._oldwndproc = None
            self._hwnd = None

    def _wndproc(self, hwnd: int, msg, wparam: int, lparam: int) -> int:
        event_name = WNDPROC_EVENT_NAMES.get(msg, msg)
        callback = self._message_map.get(msg, noop)
        vlog("_wndproc%s event name=%s, callback=%s", (hwnd, msg, wparam, lparam), event_name, callback)
        v = None
        if callback != noop:
            # run our callback
            try:
                v = callback(hwnd, msg, wparam, lparam)
                vlog("%s%s=%s", callback, (hwnd, msg, wparam, lparam), v)
            except Exception as e:
                log.error("Error: callback %s failed:", callback)
                log.estr(e)
        # if our callback doesn't define the return value, use the default handler:
        if v is None:
            v = CallWindowProcW(self._oldwndproc, hwnd, msg, wparam, lparam)
            vlog("_wndproc%s return value=%s", (hwnd, msg, wparam, lparam), v)
        return v



# This file is part of Xpra.
# Copyright (C) 2021 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence

from xpra.codecs.rgb_transform import rgb_reformat
from xpra.codecs import rgb_transform
from xpra.codecs.debug import SAVE_TO_FILE, may_save_image
from xpra.net.compression import Compressed, LevelCompressed, compressed_wrapper
from xpra.log import Logger

log = Logger("encoder", "argb")


def get_version() -> tuple[int, int]:
    return 6, 0


def get_type() -> str:
    return "rgb"


def get_encodings() -> Sequence[str]:
    return "rgb24", "rgb32"


def get_info() -> dict[str, Any]:
    return {
        "version": get_version(),
        "encodings": get_encodings(),
    }


def encode(coding: str, image, options: dict) -> tuple[str, Compressed, dict[str, Any], int, int, int, int]:
    pixel_format = image.get_pixel_format()
    # log("rgb_encode%s pixel_format=%s, rgb_formats=%s",
    #    (coding, image, rgb_formats, supports_transparency, speed, rgb_zlib, rgb_lz4), pixel_format, rgb_formats)
    if pixel_format in ("BGRX", "BGRA", "RGBA", "RGBX"):
        rgb_formats = options.get("rgb_formats", ("BGRX", "BGRA", "RGBA", "RGBX"))
    elif pixel_format in ("RGB", "BGR"):
        rgb_formats = options.get("rgb_formats", ("RGB", "BGR"))
    elif pixel_format in ("r210", "BGR565"):
        rgb_formats = options.get("rgb_formats", (pixel_format, ))
    else:
        raise ValueError(f"unsupported pixel format {pixel_format!r}")
    supports_transparency = options.get("alpha", True)
    if pixel_format not in rgb_formats:
        log("rgb_encode reformatting because %s not in %s, supports_transparency=%s",
            pixel_format, rgb_formats, supports_transparency)
        if not rgb_reformat(image, rgb_formats, supports_transparency):
            raise ValueError(f"no compatible rgb format for {pixel_format!r}! (only: {rgb_formats})")
        # get the new format:
        pixel_format = image.get_pixel_format()
        # switch encoding if necessary:
        if len(pixel_format) == 4:
            coding = "rgb32"
        elif len(pixel_format) == 3:
            coding = "rgb24"
        else:
            raise ValueError(f"invalid pixel format {pixel_format!r}")
    # we may still want to re-stride:
    image.may_restride()
    # always tell client which pixel format we are sending:
    client_options = {"rgb_format": pixel_format}

    # compress here and return a wrapper so network code knows it is already zlib compressed:
    pixels = image.get_pixels()
    if not pixels:
        raise RuntimeError(f"failed to get pixels from {image}")
    width = image.get_width()
    height = image.get_height()
    stride = image.get_rowstride()
    speed = options.get("speed", 50)

    # compression stage:
    level = 0
    size = len(pixels)
    lz4 = options.get("lz4", False)
    if lz4 and size >= 512 and speed < 100:
        if size >= 4096:
            level = 1 + max(0, min(7, int(100 - speed) // 14))
        else:
            # fewer pixels, make it more likely we won't bother compressing
            # and use a lower level (max=3)
            level = max(0, min(3, int(125 - speed) // 35))
    if level > 0:
        algo = "lz4"
        can_inline = size <= 32768
        cwrapper = compressed_wrapper(coding, pixels, level=level,
                                      lz4=lz4,
                                      can_inline=can_inline)
        if isinstance(cwrapper, LevelCompressed):
            # add compressed marker:
            client_options[cwrapper.algorithm] = cwrapper.level & 0xF
            # remove network layer compression marker
            # so that this data will be decompressed by the decode thread client side:
            cwrapper.level = 0
        elif can_inline and isinstance(pixels, memoryview):
            assert isinstance(cwrapper, Compressed)
            assert cwrapper.data == pixels
            # compression either did not work or was not enabled
            # and `memoryview` pixel data cannot be handled by the packet encoders,
            # so we have to convert it to bytes:
            cwrapper.data = rgb_transform.pixels_to_bytes(pixels)
            algo = "inlined lz4"
    else:
        # can't pass a raw buffer to rencodeplus,
        # and even if we could, the image containing those pixels may be freed by the time we get to the encoder
        algo = "not"
        cwrapper = Compressed(coding, rgb_transform.pixels_to_bytes(pixels))
    if pixel_format.find("A") >= 0 or pixel_format.find("X") >= 0:
        bpp = 32
    else:
        bpp = 24
    log("rgb_encode using level=%s for %5i bytes at %3i speed, %s compressed %4sx%-4s in %s/%s: %5s bytes down to %5s",
        level, size, speed, algo, width, height, coding, pixel_format, len(pixels), len(cwrapper.data))
    if SAVE_TO_FILE and pixel_format in ("BGRX", "BGRA", ):
        may_save_image(coding, pixels)
        from io import BytesIO
        from PIL import Image
        img = Image.frombuffer("RGB" if pixel_format == "BGRX" else "BGRA",
                               (width, height), pixels, "raw", pixel_format, stride)
        buf = BytesIO()
        img.save(buf, "png")
        data = buf.getvalue()
        may_save_image("png", data)

    # wrap it using "Compressed" so the network layer receiving it
    # won't decompress it (leave it to the client's draw thread)
    return coding, cwrapper, client_options, width, height, stride, bpp



# This file is part of Xpra.
# Copyright (C) 2020 mjharkin
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import browser_cookie3  # @UnresolvedImport

from xpra.util.str_fn import strtobytes


def get_headers(host: str, _port: int) -> dict[bytes, bytes]:     # pylint: disable=unused-argument
    cookie_domain = host
    cookie_string = ''
    # get cookies for domain and all parent domains except tld
    while cookie_domain.count('.', 2):
        cj = browser_cookie3.load(domain_name=cookie_domain)
        for c in cj:
            cookie = c.name + "=" + c.value + "; "
            # add if cookie doesn't already exist from subdomain
            if c.name + "=" not in cookie_string:
                cookie_string += cookie
        cookie_domain = cookie_domain.split('.', 1)[1]
    return {
        b"Cookie": strtobytes(cookie_string),
    }



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.screen import log_screen_sizes
from xpra.util.str_fn import bytestostr
from xpra.util.env import OSEnvContext
from xpra.net.common import PacketType
from xpra.util.version import parse_version, dict_version_trim
from xpra.scripts.config import FALSE_OPTIONS, TRUE_OPTIONS
from xpra.common import get_refresh_rate_for_value, FULL_INFO
from xpra.server.mixins.stub_server_mixin import StubServerMixin
from xpra.log import Logger

GLib = gi_import("GLib")

log = Logger("screen")
gllog = Logger("opengl")


def run_opengl_probe(cmd: list[str], env: dict[str, str], display_name: str):
    props: dict[str, Any] = {}
    try:
        # pylint: disable=import-outside-toplevel
        from subprocess import Popen, PIPE
        # we want the output so we can parse it:
        env["XPRA_REDIRECT_OUTPUT"] = "0"
        gllog(f"query_opengl() using {cmd=}, {env=}")
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=env)
        out, err = proc.communicate()
        gllog("out(%s)=%s", cmd, out)
        gllog("err(%s)=%s", cmd, err)
        if proc.returncode == 0:
            # parse output:
            for line in out.splitlines():
                parts = bytestostr(line).split("=")
                if len(parts) != 2:
                    continue
                k = parts[0].strip()
                v = parts[1].strip()
                if k in ("GLX", "GLU.version", "opengl", "pyopengl", "accelerate", "shading-language-version"):
                    props[k] = parse_version(v)
                else:
                    props[k] = v
            gllog("opengl props=%s", props)
            if props:
                glprops = typedict(props)
                if glprops.strget("success").lower() in TRUE_OPTIONS:
                    gllog.info(f"OpenGL is supported on display {display_name!r}")
                    renderer = glprops.strget("renderer").split(";")[0]
                    if renderer:
                        gllog.info(f" using {renderer!r} renderer")
                else:
                    gllog.info("OpenGL is not supported on this display")
                    probe_err = glprops.strget("error")
                    if probe_err:
                        gllog.info(f" {probe_err}")
            else:
                gllog.info("No OpenGL information available")
        else:
            props["error-details"] = bytestostr(err).strip("\n\r")
            error = "unknown error"
            for x in str(err).splitlines():
                if x.startswith("RuntimeError: "):
                    error = x[len("RuntimeError: "):]
                    break
                if x.startswith("ImportError: "):
                    error = x[len("ImportError: "):]
                    break
            props["error"] = error
            log.warn("Warning: OpenGL support check failed:")
            log.warn(f" {error}")
    except Exception as e:
        gllog("query_opengl()", exc_info=True)
        gllog.error("Error: OpenGL support check failed")
        gllog.error(f" {e!r}")
        props["error"] = str(e)
    gllog("OpenGL: %s", props)
    return props


class DisplayManager(StubServerMixin):
    """
    Mixin for servers that handle displays.
    """
    DEFAULT_REFRESH_RATE = 0
    PREFIX = "display"

    def __init__(self):
        self.randr = False
        self.bell = False
        self.default_dpi = 96
        self.bit_depth = 24
        self.dpi = 0
        self.xdpi = 0
        self.ydpi = 0
        self.antialias: dict[str, Any] = {}
        self.double_click_time = -1
        self.double_click_distance = -1, -1
        self.opengl = "no"
        self.opengl_props: dict[str, Any] = {}
        self.refresh_rate = "auto"
        self.original_desktop_display = None

    def init(self, opts) -> None:
        self.opengl = opts.opengl
        self.bell = opts.bell
        self.default_dpi = int(opts.dpi)
        self.bit_depth = self.get_display_bit_depth()
        self.refresh_rate = opts.refresh_rate

    def print_screen_info(self) -> None:
        display = os.environ.get("DISPLAY")
        if display and display.startswith(":"):
            extra = ""
            bit_depth = self.get_display_bit_depth()
            if bit_depth:
                extra = f" with {bit_depth} bit colors"
            log.info(f" connected to X11 display {display}{extra}")

    def get_display_bit_depth(self) -> int:
        return 0

    def get_refresh_rate_for_value(self, invalue) -> int:
        return get_refresh_rate_for_value(self.refresh_rate, invalue)

    def parse_hello(self, ss, caps, send_ui: bool):
        if send_ui:
            self.parse_screen_info(ss)

    def add_new_client(self, ss, c: typedict, send_ui: bool, share_count: int) -> None:
        if not send_ui:
            return
        # a bit of explanation:
        # normally these things are synchronized using xsettings, which we handle already,
        # but non-posix clients have no such thing,
        # and we don't want to expose that as an interface
        # (it's not very nice, and it is very X11 specific)
        # also, clients may want to override what is in their xsettings..
        # so if the client specifies what it wants to use, we patch the xsettings with it
        # (the actual xsettings part is done in `update_all_server_settings` in the X11 specific subclasses)
        if share_count > 0:
            log.info("sharing with %s other client(s)", share_count)
            self.dpi = 0
            self.xdpi = 0
            self.ydpi = 0
            self.double_click_time = -1
            self.double_click_distance = -1, -1
            self.antialias = {}
        else:
            dpi_caps = c.get("dpi")
            if isinstance(dpi_caps, int):
                # legacy mode, ie: html5 client
                self.dpi = self.xpdi = self.ydpi = int(dpi_caps)
            else:
                tdpi = typedict(c.dictget("dpi") or {})
                self.dpi = tdpi.intget("", 0)
                self.xdpi = tdpi.intget("x", self.xdpi)
                self.ydpi = tdpi.intget("y", self.ydpi)
            self.double_click_time = c.intget("double_click.time", -1)
            self.double_click_distance = c.intpair("double_click.distance", (-1, -1))
            self.antialias = c.dictget("antialias", {})
        log("dpi=%s, dpi.x=%s, dpi.y=%s, antialias=%s",
            self.dpi, self.xdpi, self.ydpi, self.antialias)
        log("double-click time=%s, distance=%s", self.double_click_time, self.double_click_distance)
        # if we're not sharing, reset all the settings:
        reset = share_count == 0
        self.update_all_server_settings(reset)

    def last_client_exited(self) -> None:
        self.reset_icc_profile()

    def threaded_setup(self) -> None:
        self.opengl_props = self.query_opengl()

    def query_opengl(self) -> dict[str, Any]:
        props: dict[str, Any] = {}
        if self.opengl.lower() == "noprobe" or self.opengl.lower() in FALSE_OPTIONS:
            gllog("query_opengl() skipped because opengl=%s", self.opengl)
            return props
        with OSEnvContext(XPRA_VERIFY_MAIN_THREAD="0"):
            try:
                # import OpenGL directly
                import OpenGL
                assert OpenGL
                gllog("found pyopengl version %s", OpenGL.__version__)
                # this may trigger an `AttributeError` if libGLX / libOpenGL are not installed:
                from OpenGL import GL
                assert GL
                gllog("loaded `GL` bindings")
            except (ImportError, AttributeError) as e:
                return {
                    'error': f'OpenGL is not available: {e}',
                    'success': False,
                }
            try:
                from xpra.opengl import backing
                assert backing
            except ImportError:
                return {
                    'error': '`xpra.opengl` is not available',
                    'success': False,
                }
        from xpra.platform.paths import get_xpra_command
        cmd = self.get_full_child_command(get_xpra_command() + ["opengl", "--opengl=force"])
        return run_opengl_probe(cmd, self.get_child_env(), self.display_name)

    def get_caps(self, source) -> dict[str, Any]:
        caps: dict[str, Any] = {
            "bell": self.bell,
        }
        if "display" in source.wants:
            root_size = self.get_root_window_size()
            if root_size:
                caps |= {
                    "actual_desktop_size": root_size,
                    "root_window_size": root_size,
                    "desktop_size": self._get_desktop_size_capability(source, *root_size),
                }
        if FULL_INFO and self.opengl_props:
            caps["opengl"] = dict_version_trim(self.opengl_props)
        return caps

    def get_server_features(self, source) -> dict[str, Any]:
        features: dict[str, Any] = {}
        if source and "display" in source.wants:
            max_size = self.get_max_screen_size()
            if max_size:
                features["max_desktop_size"] = max_size
            display = os.environ.get("DISPLAY")
            if display:
                features["display"] = display
        return features

    def get_ui_info(self, proto, client_uuids=None, *args) -> dict[str, Any]:
        max_size = self.get_max_screen_size()
        if max_size:
            return {"server": {"max_desktop_size": max_size}}
        return {}

    def get_info(self, _proto) -> dict[str, Any]:
        i = {
            "randr": self.randr,
            "bell": self.bell,
            "double-click": {
                "time": self.double_click_time,
                "distance": self.double_click_distance,
            },
            "dpi": {
                "default": self.default_dpi,
                "value": self.dpi,
                "x": self.xdpi,
                "y": self.ydpi,
            },
            "antialias": self.antialias,
            "depth": self.bit_depth,
            "refresh-rate": self.refresh_rate,
        }
        if self.original_desktop_display:
            i["original-desktop-display"] = self.original_desktop_display
        if self.opengl_props:
            i["opengl"] = self.opengl_props
        return {
            "display": i,
        }

    def _process_set_bell(self, proto, packet: PacketType) -> None:
        assert self.bell, "cannot toggle send_bell: the feature is disabled"
        ss = self.get_server_source(proto)
        if ss:
            ss.send_bell = bool(packet[1])

    ######################################################################
    # display / screen / root window:
    def set_screen_geometry_attributes(self, w: int, h: int) -> None:
        # by default, use the screen as desktop area:
        self.set_desktop_geometry_attributes(w, h)

    def set_desktop_geometry_attributes(self, w: int, h: int) -> None:
        self.calculate_desktops()
        self.calculate_workarea(w, h)
        self.set_desktop_geometry(w, h)

    def parse_screen_info(self, ss) -> tuple[int, int]:
        return self.do_parse_screen_info(ss, ss.desktop_size)

    def do_parse_screen_info(self, ss, desktop_size) -> tuple[int, int]:
        log("do_parse_screen_info%s", (ss, desktop_size))
        dw, dh = None, None
        if desktop_size:
            try:
                dw, dh = desktop_size
                log.info(" client root window size is %sx%s", dw, dh)
                if ss.screen_sizes:
                    log_screen_sizes(dw, dh, ss.screen_sizes)
            except Exception:
                dw, dh = None, None
        best = self.configure_best_screen_size()
        if not best:
            return desktop_size
        sw, sh = best
        # we will tell the client about the size chosen in the hello we send back,
        # so record this size as the current server desktop size to avoid change notifications:
        ss.desktop_size_server = sw, sh
        # prefer desktop size, fallback to screen size:
        w = dw or sw
        h = dh or sh
        # clamp to max supported:
        max_size = self.get_max_screen_size()
        if max_size:
            maxw, maxh = max_size
            w = min(w, maxw)
            h = min(h, maxh)
        self.set_desktop_geometry_attributes(w, h)
        self.set_icc_profile()
        self.apply_refresh_rate(ss)
        log("configure_best_screen_size()=%s", (w, h))
        return w, h

    def set_icc_profile(self) -> None:
        log("set_icc_profile() not implemented")

    def reset_icc_profile(self) -> None:
        log("reset_icc_profile() not implemented")

    def _monitors_changed(self, screen) -> None:
        self.do_screen_changed(screen)

    def _screen_size_changed(self, screen) -> None:
        self.do_screen_changed(screen)

    def do_screen_changed(self, screen) -> None:
        log("do_screen_changed(%s)", screen)
        # randr has resized the screen, tell the client (if it supports it)
        w, h = screen.get_width(), screen.get_height()
        log("new screen dimensions: %ix%i", w, h)
        self.set_screen_geometry_attributes(w, h)
        self.notify_screen_changed(screen)

    def notify_screen_changed(self, screen) -> None:
        GLib.idle_add(self.send_updated_screen_size)

    def get_root_window_size(self) -> tuple[int, int]:
        raise NotImplementedError()

    def send_updated_screen_size(self) -> None:
        root_size = self.get_root_window_size()
        if not root_size:
            return
        root_w, root_h = root_size
        max_size = self.get_max_screen_size()
        if not max_size:
            return
        max_w, max_h = max_size
        root_w = min(root_w, max_w)
        root_h = min(root_h, max_h)
        count = 0
        for ss in self._server_sources.values():
            if ss.updated_desktop_size(root_w, root_h, max_w, max_h):
                count += 1
        if count > 0:
            log.info("sent updated screen size to %s clients: %sx%s (max %sx%s)",
                     count, root_w, root_h, max_w, max_h)

    def get_max_screen_size(self) -> tuple[int, int]:
        return self.get_root_window_size()

    def _get_desktop_size_capability(self, server_source, root_w: int, root_h: int) -> tuple[int, int]:
        client_size = server_source.desktop_size
        log("client resolution is %s, current server resolution is %sx%s", client_size, root_w, root_h)
        if not client_size:
            # client did not specify size, just return what we have
            return root_w, root_h
        client_w, client_h = client_size
        w = min(client_w, root_w)
        h = min(client_h, root_h)
        return w, h

    def configure_best_screen_size(self) -> tuple[int, int]:
        return self.get_root_window_size()

    def apply_refresh_rate(self, ss) -> int:
        rrate = self.get_client_refresh_rate(ss)
        log(f"apply_refresh_rate({ss}) rate={rrate}")
        if rrate > 0:
            self.set_window_refresh_rate(ss, rrate)
        return rrate

    def set_window_refresh_rate(self, ss, rrate: int):
        if hasattr(ss, "default_batch_config"):
            ss.default_batch_config.match_vrefresh(rrate)
        if hasattr(ss, "global_batch_config"):
            ss.global_batch_config.match_vrefresh(rrate)
        if hasattr(ss, "all_window_sources"):
            for window_source in ss.all_window_sources():
                bc = window_source.batch_config
                if bc:
                    bc.match_vrefresh(rrate)

    def get_client_refresh_rate(self, ss) -> int:
        vrefresh = []
        # use the refresh-rate value from the monitors
        # (value is pre-multiplied by 1000!)
        if ss.monitors:
            for mdef in ss.monitors.values():
                v = mdef.get("refresh-rate", 0)
                if v:
                    vrefresh.append(v)
        if not vrefresh and getattr(ss, "vrefresh", 0) > 0:
            vrefresh.append(ss.vrefresh * 1000)
        if not vrefresh:
            vrefresh.append(self.DEFAULT_REFRESH_RATE)
        rrate = 0
        if vrefresh:
            rrate = min(vrefresh)
            if self.refresh_rate:
                rrate = get_refresh_rate_for_value(self.refresh_rate, rrate, multiplier=1000)
            rrate //= 1000
        log("get_client_refresh_rate(%s)=%s (from %s)", ss, rrate, vrefresh)
        return rrate

    def _process_desktop_size(self, proto, packet: PacketType) -> None:
        log("new desktop size from %s: %s", proto, packet)
        ss = self.get_server_source(proto)
        if ss is None:
            return
        width, height = packet[1:3]
        ss.desktop_size = (width, height)
        if len(packet) >= 12:
            ss.set_monitors(packet[11])
        elif len(packet) >= 11:
            # fallback to the older global attribute:
            v = packet[10]
            if 0 < v < 240 and hasattr(ss, "vrefresh") and getattr(ss, "vrefresh") != v:
                ss.vrefresh = v
        if len(packet) >= 10:
            # added in 0.16 for scaled client displays:
            xdpi, ydpi = packet[8:10]
            if xdpi != self.xdpi or ydpi != self.ydpi:
                self.xdpi, self.ydpi = xdpi, ydpi
                log("new dpi: %ix%i", self.xdpi, self.ydpi)
                self.dpi = round((self.xdpi + self.ydpi) / 2)
                self.dpi_changed()
        if len(packet) >= 8:
            # added in 0.16 for scaled client displays:
            ss.desktop_size_unscaled = packet[6:8]
        if len(packet) >= 6:
            desktops, desktop_names = packet[4:6]
            ss.set_desktops(desktops, desktop_names)
            self.calculate_desktops()
        if len(packet) >= 4:
            ss.set_screen_sizes(packet[3])
        log("client requesting new size: %sx%s", width, height)
        self.set_screen_size(width, height)
        self.set_desktop_geometry_attributes(width, height)
        if len(packet) >= 4:
            log.info("received updated display dimensions")
            log.info("client display size is %sx%s", width, height)
            log_screen_sizes(width, height, ss.screen_sizes)
            self.calculate_workarea(width, height)
        self.apply_refresh_rate(ss)
        # ensures that DPI and antialias information gets reset:
        self.update_all_server_settings()

    def _process_configure_display(self, proto, packet: PacketType) -> None:
        ss = self.get_server_source(proto)
        if ss is None:
            return
        attrs = typedict(packet[1])
        desktop_size = attrs.intpair("desktop-size")
        if desktop_size:
            ss.desktop_size = desktop_size
        desktop_size_unscaled = attrs.intpair("desktop-size-unscaled")
        if desktop_size_unscaled:
            ss.desktop_size_unscaled = desktop_size_unscaled
        # vrefresh may be overridden in 'monitors' data:
        vrefresh = attrs.intget("vrefresh")
        if 0 < vrefresh < 240 and hasattr(ss, "vrefresh") and getattr(ss, "vrefresh") != vrefresh:
            ss.vrefresh = vrefresh
        monitors = attrs.dictget("monitors")
        if monitors:
            ss.set_monitors(monitors)
        if desktop_size:
            width, height = desktop_size
            log("client requesting new size: %sx%s", width, height)
            self.set_screen_size(width, height)
            log.info("received updated display dimensions")
            log.info(f"client display size is {width}x{height}")
            log_screen_sizes(width, height, ss.screen_sizes)
            self.calculate_workarea(width, height)
            self.set_desktop_geometry_attributes(width, height)
        # DPI
        dpi = 0
        dpi_caps = attrs.get("dpi")
        # unprefixed legacy mode:
        if isinstance(dpi_caps, int):
            dpi = int(dpi_caps)
        dpix = attrs.intget("dpi.x", dpi)
        dpiy = attrs.intget("dpi.y", dpi)
        # namespaced caps:
        if isinstance(dpi_caps, dict):
            tdpi = typedict(dpi_caps)
            dpix = tdpi.intget("x", dpix)
            dpiy = tdpi.intget("y", dpiy)
        if dpix and dpiy and (dpix != self.xdpi or dpiy != self.ydpi):
            self.xdpi, self.ydpi = dpix, dpiy
            log("new dpi: %ix%i", dpix, dpiy)
            self.dpi = round((dpix + dpiy) / 2)
            self.dpi_changed()
        desktop_names = attrs.strtupleget("desktop-names")
        if desktop_names:
            ss.set_desktops(attrs.intget("desktops", len(desktop_names)), desktop_names)
            self.calculate_desktops()
        iccd = attrs.dictget("icc")
        if iccd:
            iccd = typedict(iccd)
            ss.icc = iccd.get("global", ss.icc)
            ss.display_icc = iccd.get("display", ss.display_icc)
            self.set_icc_profile()
        self.apply_refresh_rate(ss)
        # ensures that DPI and antialias information gets reset:
        self.update_all_server_settings()

    def dpi_changed(self) -> None:
        """
        The x11 servers override this method
        to also update the XSettings.
        """

    def calculate_desktops(self):
        """ seamless servers can update the desktops """

    def calculate_workarea(self, w: int, h: int):
        raise NotImplementedError()

    def set_workarea(self, workarea) -> None:
        pass

    ######################################################################
    # screenshots:
    def _process_screenshot(self, proto, _packet: PacketType) -> None:
        packet = self.make_screenshot_packet()
        ss = self.get_server_source(proto)
        if packet and ss:
            ss.send(*packet)

    def make_screenshot_packet(self):
        with log.trap_error("Error making screenshot packet"):
            return self.do_make_screenshot_packet()

    def do_make_screenshot_packet(self):
        raise NotImplementedError("no screenshot capability in %s" % type(self))

    def send_screenshot(self, proto) -> None:
        # this is a screenshot request, handle it and disconnect
        try:
            packet = self.make_screenshot_packet()
            if not packet:
                self.send_disconnect(proto, "screenshot failed")
                return
            proto.send_now(packet)
            GLib.timeout_add(5 * 1000, self.send_disconnect, proto, "screenshot sent")
        except Exception as e:
            log.error("failed to capture screenshot", exc_info=True)
            self.send_disconnect(proto, "screenshot failed: %s" % e)

    def init_packet_handlers(self) -> None:
        self.add_packets("set-bell", "desktop_size", "configure-display", "screenshot", main_thread=True)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import platform
import time
import socket
import sys

def measure(cb):
    start = time.time()
    v = cb()
    print("%s()=%s" % (cb, v))
    end = time.time()
    elapsed = 1000*(end-start)
    if elapsed>1000:
        sys.exit(1)
    print("elapsed time: %sms" % int(elapsed))

def main():
    measure(platform.node)
    measure(socket.gethostname)
    measure(socket.getfqdn)

if __name__ == "__main__":
    main()



# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from math import sqrt
from io import BytesIO
from typing import Any
from collections.abc import Sequence

import PIL
from PIL import Image, ImagePalette, __version__ as pil_version

from xpra.common import roundup
from xpra.codecs.debug import may_save_image
from xpra.codecs.image import ImageWrapper
from xpra.util.objects import typedict
from xpra.util.str_fn import csv, hexstr
from xpra.net.compression import Compressed
from xpra.log import Logger

log = Logger("encoder", "pillow")

ENCODE_FORMATS = os.environ.get("XPRA_PILLOW_ENCODE_FORMATS", "png,png/L,png/P,jpeg,webp").split(",")

Image.init()

try:
    pil_major = int(pil_version.split(".")[0])
except ValueError:
    pil_major = 0


try:
    # pylint: disable=ungrouped-imports
    from PIL.Image import Palette, Resampling

    ADAPTIVE = Palette.ADAPTIVE
    WEB = Palette.WEB
    NEAREST = Resampling.NEAREST
    BILINEAR = Resampling.BILINEAR
    BICUBIC = Resampling.BICUBIC
    LANCZOS = Resampling.LANCZOS
except ImportError:
    # location for older versions:
    from PIL.Image import ADAPTIVE, WEB  # type: ignore
    from PIL.Image import NEAREST, BILINEAR, BICUBIC, LANCZOS  # type: ignore


def get_version() -> str:
    return PIL.__version__


def get_type() -> str:
    return "pillow"


def do_get_encodings() -> Sequence[str]:
    log("PIL.Image.SAVE=%s", Image.SAVE)
    encodings: list[str] = []
    for encoding in ENCODE_FORMATS:
        # strip suffix (so "png/L" -> "png")
        stripped = encoding.split("/")[0].upper()
        if stripped in Image.SAVE:
            encodings.append(encoding)
    log("do_get_encodings()=%s", encodings)
    return tuple(encodings)


def get_encodings() -> Sequence[str]:
    return ENCODINGS


ENCODINGS: Sequence[str] = do_get_encodings()


def get_info() -> dict[str, Any]:
    return {
        "version": get_version(),
        "encodings": get_encodings(),
    }


def encode(coding: str, image: ImageWrapper, options: typedict) -> tuple[str, Compressed, dict[str, Any], int, int, int, int]:
    if coding not in ("jpeg", "webp", "png", "png/P", "png/L"):
        raise ValueError(f"unsupported encoding: {coding!r}")
    log("pillow.encode%s", (coding, image, options))
    quality = options.intget("quality", 50)
    speed = options.intget("speed", 50)
    supports_transparency = options.boolget("alpha", True)
    grayscale = options.boolget("grayscale", False)
    pixel_format: str = image.get_pixel_format()
    palette: list[int] = []
    w: int = image.get_width()
    h: int = image.get_height()
    bpp = 32
    in_stride = image.get_rowstride()
    rowstride = in_stride
    pixels = image.get_pixels()
    if not pixels:
        raise RuntimeError(f"failed to get pixels from {image}")
    # remove transparency if it cannot be handled,
    # and deal with non-24-bit formats:
    if pixel_format == "r210":
        from xpra.codecs.argb.argb import r210_to_rgba, r210_to_rgb  # pylint: disable=import-outside-toplevel
        has_alpha = options.intget("depth", 24) == 32
        if has_alpha and supports_transparency:
            rowstride = w * 4
            pixels = r210_to_rgba(pixels, w, h, in_stride, w * 4)
            pixel_format = "RGBA"
        else:
            rowstride = roundup(w * 3, 2)
            pixels = r210_to_rgb(pixels, w, h, in_stride, rowstride)
            pixel_format = "RGB"
            bpp = 24
    elif pixel_format == "BGR565":
        from xpra.codecs.argb.argb import bgr565_to_rgbx, bgr565_to_rgb  # pylint: disable=import-outside-toplevel
        if supports_transparency:
            rowstride = in_stride * 2
            pixels = bgr565_to_rgbx(pixels)
            pixel_format = "RGBA"
        else:
            rowstride = in_stride * 3 // 2
            pixels = bgr565_to_rgb(pixels)
            pixel_format = "RGB"
            bpp = 24
    elif pixel_format == "RLE8":
        pixel_format = "P"
        palette = []
        # pillow requires 8 bit palette values,
        # but we get 16-bit values from the image wrapper (X11 palettes are 16-bit):
        for r, g, b in image.get_palette():
            palette.append((r >> 8) & 0xFF)
            palette.append((g >> 8) & 0xFF)
            palette.append((b >> 8) & 0xFF)
        bpp = 8
    else:
        if pixel_format not in ("RGBA", "RGBX", "BGRA", "BGRX", "BGR", "RGB"):
            raise ValueError(f"invalid pixel format {pixel_format!r}")
    rgb: str = {
        "XRGB": "RGB",
        "BGRX": "RGB",
        "RGBX": "RGB",
        "RGBA": "RGBA",
        "BGRA": "RGBA",
        "BGR": "RGB",
    }.get(pixel_format, pixel_format)
    pil_import_format = pixel_format.replace("A", "a")
    try:
        # older versions cannot use the memoryview directly:
        if pil_major < 10 and isinstance(pixels, memoryview):
            pixels = pixels.tobytes()
        # it is safe to use frombuffer() here since the convert()
        # calls below will not convert and modify the data in place,
        # and we save the compressed data then discard the image
        im = Image.frombuffer(rgb, (w, h), pixels, "raw", pil_import_format, rowstride, 1)
    except Exception as e:
        log("Image.frombuffer%s", (
            rgb, (w, h), len(pixels),
            "raw", pil_import_format, rowstride, 1),
            exc_info=True)
        log.error(f"Error: pillow failed to import image as {pil_import_format!r}:")
        log.estr(e)
        log.error(" for %s", image)
        log.error(f" pixel data: %i %s, with {rowstride=}", len(pixels), type(pixels))
        raise
    try:
        if palette:
            im.putpalette(palette)
            im.palette = ImagePalette.ImagePalette("RGB", palette=palette)
        if coding != "png/L" and grayscale:
            if rgb.find("A") >= 0 and supports_transparency and coding != "jpeg":
                im = im.convert("LA")
            else:
                im = im.convert("L")
            rgb = "L"
            bpp = 8
        elif coding.startswith("png") and not supports_transparency and rgb == "RGBA":
            im = im.convert("RGB")
            rgb = "RGB"
            bpp = 24
    except Exception as e:
        log.error("Error: pillow failed to convert image")
        log.estr(e)
        log.error(" for %s", im)
        raise
    scaled_width = options.intget("scaled-width", w)
    scaled_height = options.intget("scaled-height", h)
    client_options: dict[str, Any] = {}
    if scaled_width != w or scaled_height != h:
        if speed >= 95:
            resample = NEAREST
        elif speed > 80:
            resample = BILINEAR
        elif speed >= 30:
            resample = BICUBIC
        else:
            resample = LANCZOS
        im = im.resize((scaled_width, scaled_height), resample=resample)
        client_options["resample"] = getattr(resample, "name", str(resample))
    if coding in ("jpeg", "webp"):
        # newer versions of pillow require explicit conversion to non-alpha:
        if pixel_format.find("A") >= 0 and coding == "jpeg":
            im = im.convert("RGB")
        q = int(min(100, max(1, quality)))
        kwargs = dict(im.info)
        kwargs["quality"] = q
        if coding == "webp":
            kwargs["method"] = int(speed < 10)
            client_options["quality"] = min(100, max(0, 10 + round(sqrt(sqrt(q * 100) * 100))))
        else:
            client_options["quality"] = min(99, q)
        if coding == "jpeg" and speed < 50:
            # (optimizing jpeg is pretty cheap and worth doing)
            kwargs["optimize"] = True
        elif coding == "webp" and q >= 100:
            kwargs["lossless"] = True
            kwargs["quality"] = 0
        pil_fmt = coding.upper()
    else:
        assert coding in ("png", "png/P", "png/L"), "unsupported encoding: %r" % coding
        if coding in ("png/L", "png/P") and supports_transparency and rgb == "RGBA":
            # grab alpha channel (the last one):
            # we use the last channel because we know it is RGBA,
            # otherwise we should do: alpha_index= image.getbands().index('A')
            alpha = im.split()[-1]

            # convert to simple on or off mask:
            # set all pixel values below 128 to 255, and the rest to 0

            def mask_value(a) -> int:
                # without branch, same as:
                # return 255 if a <= 128 else 0
                return int(a <= 128) * 255

            mask = Image.eval(alpha, mask_value)
        else:
            # no transparency
            mask = None
        if coding == "png/L":
            im = im.convert("L", palette=ADAPTIVE, colors=255)
            bpp = 8
        elif coding == "png/P":
            # convert to 255 indexed colour if:
            # * we're not in palette mode yet (source is >8bpp)
            # * we need space for the mask (256 -> 255)
            if mask or not palette:
                # I wanted to use the "better" adaptive method,
                # but this does NOT work (produces a black image instead):
                # im.convert("P", palette=Image.ADAPTIVE)
                im = im.convert("P", palette=WEB, colors=255)
            bpp = 8
        kwargs = im.info
        if mask:
            # paste the alpha mask to the color of index 255
            im.paste(255, mask)
            client_options["transparency"] = 255
            kwargs["transparency"] = 255
        if speed == 0:
            # optimizing png is very rarely worth doing
            kwargs["optimize"] = True
        # level can range from 0 to 9, but anything above 5 is way too slow for small gains:
        # 76-100   -> 1
        # 51-76    -> 2
        # etc
        level = max(1, min(5, (100 - speed) // 25))
        kwargs["compress_level"] = level
        # no need to expose to the client:
        # client_options["compress_level"] = level
        # default is good enough, no need to override, other options:
        # DEFAULT_STRATEGY, FILTERED, HUFFMAN_ONLY, RLE, FIXED
        # kwargs["compress_type"] = Image.DEFAULT_STRATEGY
        pil_fmt = "PNG"
    buf = BytesIO()
    im.save(buf, pil_fmt, **kwargs)
    data = buf.getvalue()
    may_save_image(pil_fmt, data)
    log("sending %sx%s %s as %s, mode=%s, options=%s", w, h, pixel_format, coding, im.mode, kwargs)
    buf.close()
    return coding, Compressed(coding, data), client_options, image.get_width(), image.get_height(), 0, bpp


def selftest(full=False) -> None:
    global ENCODINGS
    # pylint: disable=import-outside-toplevel
    from xpra.codecs.checks import make_test_image
    img = make_test_image("BGRA", 128, 128)
    vrange: Sequence[int] = (50,)
    if full:
        vrange = (0, 50, 100)
    for encoding in tuple(ENCODINGS):
        q = s = 0
        alpha = False
        try:
            for q in vrange:
                for s in vrange:
                    for alpha in (True, False):
                        v = encode(encoding, img, typedict({
                            "quality": q,
                            "speed": s,
                            "alpha": alpha,
                        }))
                        assert v, "encode output was empty!"
                        cdata = v[1].data
                        log("encode(%s)=%s", (encoding, img, q, s, alpha), hexstr(cdata))
        except Exception as e:  # pragma: no cover
            log_fn = log.warn
            log_fn("Pillow error saving %s with quality=%s, speed=%s, alpha=%s", encoding, q, s, alpha)
            log_fn(" %s", e, exc_info=True)
            encs: list[str] = list(ENCODINGS)
            encs.remove(encoding)
            ENCODINGS = tuple(encs)


if __name__ == "__main__":
    selftest(True)
    print(csv(get_encodings()))



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=import-outside-toplevel


def get_client_protocol_class(socktype: str):
    if socktype in ("ws", "wss", "quic"):
        from xpra.net.websockets.protocol import WebSocketProtocol
        return WebSocketProtocol
    if socktype == "vnc":
        from xpra.client.base.rfb_protocol import RFBClientProtocol
        return RFBClientProtocol
    from xpra.net.protocol.socket_handler import SocketProtocol
    return SocketProtocol


def get_server_protocol_class(socktype: str):
    if socktype in ("ws", "wss"):
        from xpra.net.websockets.protocol import WebSocketProtocol
        return WebSocketProtocol
    if socktype == "vnc":
        from xpra.server.rfb.protocol import RFBServerProtocol
        return RFBServerProtocol
    from xpra.net.protocol.socket_handler import SocketProtocol
    return SocketProtocol



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import time
import logging
import unittest
import subprocess

from xpra.util.env import OSEnvContext
from xpra.util import child_reaper
from xpra.util.child_reaper import getChildReaper, reaper_cleanup, log


class TestChildReaper(unittest.TestCase):

    def test_childreaper(self):
        for polling in (True, False):
            with OSEnvContext():
                os.environ["XPRA_USE_PROCESS_POLLING"] = str(int(polling))
                self.do_test_child_reaper()


    def do_test_child_reaper(self):
        #force reset singleton:
        child_reaper.singleton = None
        #no-op:
        reaper_cleanup()

        log.setLevel(logging.ERROR)
        cr = getChildReaper()
        #one that exits before we add the process, one that takes longer:
        TEST_CHILDREN = (["echo"], ["sleep", "0.5"])
        count = 0
        for cmd in TEST_CHILDREN:
            cmd_info = " ".join(cmd)
            proc = subprocess.Popen(cmd)
            cr.add_process(proc, cmd_info, cmd_info, False, False, None)
            count += 1
            for _ in range(10):
                if not cr.check():
                    break
                time.sleep(0.1)
            #we can't check the returncode because it may not be set yet!
            #assert proc.poll() is not None, "%s process did not terminate?" % cmd_info
            assert cr.check() is False, "reaper did not notice that the '%s' process has terminated" % cmd_info
            i = cr.get_info()
            children = i.get("children").get("total")
            assert children==count, "expected %s children recorded, but got %s" % (count, children)

        #now check for the forget option:
        proc = subprocess.Popen(["sleep", "60"])
        procinfo = cr.add_process(proc, "sleep 60", "sleep 60", False, True, None)
        assert repr(procinfo)
        count +=1
        assert cr.check() is True, "sleep process terminated too quickly"
        i = cr.get_info()
        children = i.get("children").get("total")
        assert children==count, "expected %s children recorded, but got %s" % (count, children)
        #trying to claim it is dead when it is not:
        #(this will print some warnings)
        cr.add_dead_pid(proc.pid)
        proc.terminate()
        #now wait for the sleep process to exit:
        for _ in range(10):
            if proc.poll() is not None:
                break
            time.sleep(0.1)
        assert proc.poll() is not None
        assert cr.check() is False, "sleep process did not terminate?"
        count -= 1
        i = cr.get_info()
        children = i.get("children").get("total")
        if children!=count:
            raise Exception("expected the sleep process to have been forgotten (%s children)" % count +
            "but got %s children instead in the reaper records" % children)
        reaper_cleanup()
        #can run again:
        reaper_cleanup()
        #nothing for an invalid pid:
        assert cr.get_proc_info(-1) is None


def main():
    from xpra.os_util import WIN32
    if not WIN32:
        unittest.main()

if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from collections.abc import Callable, Sequence

from aioquic.h3.events import HeadersReceived, H3Event
from aioquic.h3.exceptions import NoAvailablePushIDError
from aioquic.quic.packet import QuicErrorCode

from xpra.net.bytestreams import pretty_socket
from xpra.net.quic.connection import XpraQuicConnection
from xpra.net.quic.common import SERVER_NAME, http_date, binary_headers
from xpra.net.websockets.header import close_packet
from xpra.util.env import first_time
from xpra.util.str_fn import Ellipsizer, std
from xpra.log import Logger

log = Logger("quic")

# can be used to use substreams based on packet prefix: sound,webcam,draw
SUBSTREAM_PACKET_TYPES = tuple(x.strip() for x in os.environ.get(
    "XPRA_QUIC_SUBSTREAM_PACKET_TYPES",
    ""
).split(",") if x.strip())


def filterpath(path: str) -> str:
    return std(path, "-+=&;%@._/")


# SUBSTREAM_PACKET_LOSS_PCT = envint("XPRA_QUIC_SUBSTREAM_PACKET_LOSS_PCT", 0)


class ServerWebSocketConnection(XpraQuicConnection):
    def __init__(self, connection, scope: dict,
                 stream_id: int, transmit: Callable[[], None]) -> None:
        super().__init__(connection, stream_id, transmit, "", 0, "wss", info=None, options=None)
        self.scope: dict = scope
        self._packet_type_streams: dict[str, int] = {}
        self._use_substreams = bool(SUBSTREAM_PACKET_TYPES)

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        info.setdefault("quic", {})["scope"] = self.scope
        return info

    def __repr__(self):
        try:
            return f"QuicConnection({pretty_socket(self.endpoint)}, {self.stream_id})"
        except AttributeError:
            return f"ServerWebSocketConnection<{self.stream_id}>"

    def http_event_received(self, event: H3Event) -> None:
        log("ws:http_event_received(%s)", Ellipsizer(event))
        if self.closed:
            return
        if isinstance(event, HeadersReceived):
            subprotocols = self.scope.get("subprotocols", ())
            if "xpra" not in subprotocols:
                message = f"unsupported websocket subprotocols {subprotocols}"
                log.warn(f"Warning: {message}")
                self.close(QuicErrorCode.APPLICATION_ERROR, message)
                return
            log.info("websocket request at %s", self.scope.get("path", "/"))
            self.accepted = True
            self.send_accept(self.stream_id)
            self.transmit()
            return
        super().http_event_received(event)

    def send_accept(self, stream_id: int) -> None:
        self.send_headers(stream_id=stream_id, headers={
            ":status": 200,
            "server": SERVER_NAME,
            "date": http_date(),
            "sec-websocket-protocol": "xpra",
        })

    def send_close(self, code=QuicErrorCode.NO_ERROR, reason="") -> None:
        log(f"send_close({code}, {reason})")
        wscode = 1000 if code == QuicErrorCode.NO_ERROR else 4000 + int(code)
        self.send_ws_close(wscode, reason)
        super().send_close(code, reason)

    def send_ws_close(self, code: int = 1000, reason: str = "") -> None:
        if self.accepted:
            data = close_packet(code, reason)
            self.write(data, "close")
        else:
            self.send_headers(self.stream_id, headers={":status": code})
            self.transmit()

    def get_packet_stream_id(self, packet_type: str) -> int:
        if self.closed or not self._use_substreams or not packet_type:
            return self.stream_id
        if not any(packet_type.startswith(x) for x in SUBSTREAM_PACKET_TYPES):
            return self.stream_id
        # ie: "sound-data" -> "sound"
        stream_type = packet_type.split("-", 1)[0]
        stream_id = self._packet_type_streams.get(stream_type)
        if stream_id is not None:
            # already allocated substream:
            return stream_id
        # allocate a new one and record it
        # (even if it fails, so we don't retry to allocate it again and again):
        stream_id = self.allocate_new_stream_id(stream_type) or self.stream_id
        self._packet_type_streams[stream_type] = stream_id
        return stream_id

    def scopestr(self, key: str, default: str, values: Sequence[str] = ()) -> str:
        value = self.scope.get(key, default)
        if values and value not in values:
            raise ValueError(f"invalid value for {key!r}: {value!r}")
        return value

    def allocate_new_stream_id(self, stream_type: str) -> int:
        log(f"allocate_new_stream_id({stream_type!r})")
        # should use more "correct" values here
        # (we don't need those headers,
        # but the client would drop the packet without them..)
        headers = binary_headers({
            ":method": self.scopestr("method", "CONNECT", ("CONNECT", "CONNECT-UDP")),
            ":scheme": self.scopestr("scheme", "wss", ("wss", "https")),
            ":authority": self.scope.get("transport-info", {}).get("sockname", ("localhost", ))[0],
            ":path": filterpath(self.scopestr("path", "/")),
        })
        try:
            stream_id = self.connection.send_push_promise(self.stream_id, headers)
        except NoAvailablePushIDError:
            log(f"unable to allocate new stream-id using {self.stream_id} and {headers}", exc_info=True)
            if first_time("quic-no-push-id"):
                log.warn(f"Warning: unable to allocate a new stream-id for {stream_type!r}")
            else:
                # more than one error, stop trying:
                self._use_substreams = False
            return 0
        log.info(f"new stream: {stream_id} for {stream_type!r} with headers={headers}")
        self.send_headers(stream_id=stream_id, headers={
            ":status": 200,
            "substream": self.stream_id,
            "stream-type": stream_type,
        })
        return stream_id



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# taken from the code I wrote for winswitch

import avahi  # @UnresolvedImport
import dbus
from socket import if_nametoindex, if_indextoname

try:
    from dbus.exceptions import DBusException
except ImportError:
    # not available in all versions of the bindings?
    DBusException = Exception

from xpra.net.mdns import XPRA_TCP_MDNS_TYPE, XPRA_UDP_MDNS_TYPE, SHOW_INTERFACE
from xpra.dbus.common import init_system_bus
from xpra.net.net_util import get_iface
from xpra.log import Logger

log = Logger("network", "mdns")


def get_interface_index(host) -> int:
    log("get_interface_index(%s)", host)
    if host in ("0.0.0.0", "", "*", "::"):
        return avahi.IF_UNSPEC

    if not if_nametoindex:
        log.warn("Warning: cannot convert interface to index (if_nametoindex is missing)")
        log.warn(" so returning 'IF_UNSPEC', avahi will publish on ALL interfaces")
        return avahi.IF_UNSPEC

    iface = get_iface(host)
    log("get_iface(%s)=%s", host, iface)
    if not iface:
        return avahi.IF_UNSPEC

    index = if_nametoindex(iface)
    log("if_nametoindex(%s)=%s", iface, index)
    if iface is None:
        return avahi.IF_UNSPEC
    return index


class AvahiPublishers:
    """
    Aggregates a number of AvahiPublisher(s).
    This takes care of constructing the appropriate AvahiPublisher
    with the interface index and port for the given list of (host,port)s to broadcast on,
    and to convert the text dict into a TXT string.
    """

    def __init__(self, listen_on, service_name: str, service_type: str = XPRA_TCP_MDNS_TYPE, text_dict=None):
        log("AvahiPublishers%s", (listen_on, service_name, service_type, text_dict))
        self.publishers = []
        try:
            bus = init_system_bus()
        except Exception as e:
            log.warn("failed to connect to the system dbus: %s", e)
            log.warn(" either start a dbus session or disable mdns support")
            return
        for host, port in listen_on:
            iface_index = get_interface_index(host)
            log("iface_index(%s)=%s", host, iface_index)
            td = text_dict or {}
            if SHOW_INTERFACE and iface_index is not None:
                td = text_dict.copy()
                try:
                    td["iface"] = if_indextoname(iface_index)
                except OSError:
                    pass
            txt = []
            if text_dict:
                for k, v in td.items():
                    txt.append(f"{k}={v}")
            fqdn = host
            if host in ("0.0.0.0", "::"):
                fqdn = ""
            elif host:
                try:
                    import socket
                    fqdn = socket.gethostbyaddr(host)[0]
                    log("gethostbyaddr(%s)=%s", host, fqdn)
                    if fqdn.find(".") < 0:
                        fqdn = socket.getfqdn(host)
                        log("getfqdn(%s)=%s", host, fqdn)
                    if fqdn.find(".") < 0:
                        if fqdn:
                            fqdn += ".local"
                        log("cannot find a fully qualified domain name for '%s', using: %s", host, fqdn)
                except (OSError, IndexError):
                    log("failed to get hostbyaddr for '%s'", host, exc_info=True)
            self.publishers.append(AvahiPublisher(bus, service_name, port,
                                                  service_type, domain="", host=fqdn,
                                                  text=txt, interface=iface_index))

    def start(self) -> None:
        log("avahi:starting: %s", self.publishers)
        if not self.publishers:
            return
        all_err = True
        for publisher in self.publishers:
            if publisher.start():
                all_err = False
        if all_err:
            log.warn(" to avoid this warning, disable mdns support ")
            log.warn(" using the 'mdns=no' option")

    def stop(self) -> None:
        log("stopping: %s", self.publishers)
        for publisher in self.publishers:
            publisher.stop()

    def update_txt(self, txt) -> None:
        for publisher in self.publishers:
            publisher.update_txt(txt)


class AvahiPublisher:

    def __init__(self, bus, name, port, stype=XPRA_TCP_MDNS_TYPE,
                 domain="", host="", text=(), interface=avahi.IF_UNSPEC):
        log("AvahiPublisher%s", (bus, name, port, stype, domain, host, text, interface))
        self.bus = bus
        self.name = name
        self.stype = stype
        self.domain = domain
        if host == "::":
            host = ""
        self.host = host
        self.port = port
        self.text = avahi.string_array_to_txt_array(text)
        self.interface = interface
        self.server = None
        self.group = None

    def iface(self) -> str:
        if self.interface > 0:
            return "interface %i" % self.interface
        return "all interfaces"

    def host_str(self) -> str:
        return "%s %s:%s on %s" % (self.name, self.host, self.port, self.iface())

    def __repr__(self):
        return "AvahiPublisher(%s)" % self.host_str()

    def start(self) -> bool:
        try:
            self.server = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER),
                                         avahi.DBUS_INTERFACE_SERVER)
            self.group = dbus.Interface(self.bus.get_object(avahi.DBUS_NAME, self.server.EntryGroupNew()),
                                        avahi.DBUS_INTERFACE_ENTRY_GROUP)
        except Exception as e:
            log.warn("failed to connect to avahi's dbus interface: %s", e)
            return False
        log("avahi dbus server=%s", self.server)
        log("avahi dbus group=%s", self.group)
        self.server.connect_to_signal("StateChanged", self.server_state_changed)
        return self.server_state_changed(self.server.GetState())

    def server_state_changed(self, state, error=None) -> bool:
        log("server_state_changed(%s, %s) on %s", state, error, self.server)
        if state == avahi.SERVER_COLLISION:
            log.error("Error: mdns server name collision")
            if error:
                log.error(" %s", error)
            log.error(" for name '%s' and port %i on %s", self.name, self.port, self.iface())
            self.stop()
            return False
        if state == avahi.SERVER_RUNNING:
            self.add_service()
            return True
        state_str = "unknown"
        for const in ("INVALID", "REGISTERING", "COLLISION", "FAILURE"):
            val = getattr(avahi, "SERVER_%s" % const, None)
            if val is not None and val == state:
                state_str = const
                break
        log.warn("Warning: mdns server state changed to '%s'", state_str)
        log.warn(" for name '%s' and port %i on %s", self.name, self.port, self.iface())
        return False

    def add_service(self) -> None:
        if not self.group:
            return
        try:
            args = (
                self.interface, avahi.PROTO_UNSPEC, dbus.UInt32(0),
                self.name, self.stype, self.domain, self.host,
                dbus.UInt16(self.port), self.text,
            )
            log("calling %s%s", self.group, args)
            self.group.AddService(*args)
            self.group.Commit()
            log("dbus service added")
        except DBusException as e:
            log("cannot add service", exc_info=True)
            # use try+except as older versions may not have those modules?
            message = e.get_dbus_message()
            dbus_error_name = e.get_dbus_name()
            log.error("Error starting publisher %s", self.host_str())
            if dbus_error_name == "org.freedesktop.Avahi.CollisionError":
                log.error(" another instance already claims this dbus name")
                log.estr(e)
                log.error(" %s", message)
            else:
                for line in str(e).splitlines():
                    for part in line.split(":", 1):
                        if part:
                            log.error(f" {part}")
            self.stop()

    def stop(self) -> None:
        group = self.group
        log("%s.stop() group=%s", self, group)
        if group:
            self.group = None
            try:
                group.Reset()
            except Exception as e:
                log.error("Error stopping avahi publisher %s:", self)
                log.estr(e)
        self.server = None

    def update_txt(self, txt) -> None:
        if not self.server:
            log("update_txt(%s) ignored, already stopped", txt)
            return
        if not self.group:
            log.warn("Warning: cannot update mdns record")
            log.warn(" publisher has already been stopped")
            return
        # prevent avahi from choking on ints:
        txt_strs = {k: str(v) for k, v in txt.items()}

        def reply_handler(*args):
            log("reply_handler%s", args)
            log("update_txt(%s) done", txt)

        def error_handler(*args):
            log("error_handler%s", args)
            log.warn("Warning: failed to update mDNS TXT record")
            log.warn(" for name '%s'", self.name)
            log.warn(" host=%s, port=%s", self.host, self.port)
            log.warn(" with new data:")
            for k, v in txt_strs.items():
                log.warn(" * %s=%s", k, v)

        txt_array = avahi.dict_to_txt_array(txt_strs)
        self.group.UpdateServiceTxt(self.interface,
                                    avahi.PROTO_UNSPEC, dbus.UInt32(0), self.name, self.stype, self.domain,
                                    txt_array, reply_handler=reply_handler,
                                    error_handler=error_handler)


def main():
    from xpra.os_util import gi_import
    GLib = gi_import("GLib")
    import random
    import signal
    port = int(20000 * random.random()) + 10000
    host = ""
    name = "test service"
    bus = init_system_bus()
    publishers = []

    def add(service_type: str = XPRA_TCP_MDNS_TYPE):
        publisher = AvahiPublisher(bus, name, port, stype=service_type, host=host, text=("somename=somevalue",))
        publishers.append(publisher)

        def start():
            publisher.start()

        def update_rec():
            publisher.update_txt({b"hello": b"world"})

        GLib.idle_add(start)
        GLib.timeout_add(5 * 1000, update_rec)

    add(XPRA_TCP_MDNS_TYPE)
    add(XPRA_UDP_MDNS_TYPE)
    loop = GLib.MainLoop()

    def exit_loop(*_args):
        loop.quit()
    signal.signal(signal.SIGTERM, exit_loop)
    loop.run()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from ctypes import (
    POINTER, Structure, c_int, byref, cast, sizeof,
    WinDLL,  # @UnresolvedImport
)
from ctypes.wintypes import WORD, DWORD, HANDLE, BOOL, LPSTR
from xpra.util.str_fn import print_nested_dict

PDWORD = POINTER(DWORD)

wtsapi32 = WinDLL("WtsApi32")

NOTIFY_FOR_THIS_SESSION = 0

WTS_CURRENT_SERVER_HANDLE = 0

# no idea where we're supposed to get those from:
WM_WTSSESSION_CHANGE = 0x02b1
WM_DWMNCRENDERINGCHANGED = 0x031F
WM_DWMCOMPOSITIONCHANGED = 0x031E


class WTS_SESSION_INFOA(Structure):
    _fields_ = [
        ("SessionId", DWORD),
        ("pWinStationName", LPSTR),
        ("State", c_int),
    ]


class WTS_CLIENT_DISPLAY(Structure):
    _fields_ = [
        ("HorizontalResolution", DWORD),
        ("VerticalResolution", DWORD),
        ("ColorDepth", DWORD),
    ]


PWTS_SESSION_INFOA = POINTER(WTS_SESSION_INFOA)
PPWTS_SESSION_INFOA = POINTER(PWTS_SESSION_INFOA)
WTSActive = 0
WTSConnected = 1
WTSConnectQuery = 2
WTSShadow = 3
WTSDisconnected = 4
WTSIdle = 5
WTSlisten = 6
WTSReset = 7
WTSDown = 8
WTSInit = 9
CONNECT_STATE: dict[int, str] = {
    WTSActive: "Active",
    WTSConnected: "Connected",
    WTSConnectQuery: "ConnectQuery",
    WTSShadow: "Shadow",
    WTSDisconnected: "Disconnected",
    WTSIdle: "Idle",
    WTSlisten: "listen",
    WTSReset: "Reset",
    WTSDown: "Down",
    WTSInit: "Init",
}
WTSInitialProgram = 0
WTSApplicationName = 1
WTSWorkingDirectory = 2
WTSOEMId = 3
WTSSessionId = 4
WTSUserName = 5
WTSWinStationName = 6
WTSDomainName = 7
WTSConnectState = 8
WTSClientBuildNumber = 9
WTSClientName = 10
WTSClientDirectory = 11
WTSClientProductId = 12
WTSClientHardwareId = 13
WTSClientAddress = 14
WTSClientDisplay = 15
WTSClientProtocolType = 16
WTSIdleTime = 17
WTSLogonTime = 18
WTSIncomingBytes = 19
WTSOutgoingBytes = 20
WTSIncomingFrames = 21
WTSOutgoingFrames = 22
WTSClientInfo = 23
WTSSessionInfo = 24
WTSSessionInfoEx = 25
WTSConfigInfo = 26
WTSValidationInfo = 27
WTSSessionAddressV4 = 28
WTSIsRemoteSession = 29
WTS_INFO_CLASS: dict[int, str] = {
    WTSInitialProgram: "InitialProgram",
    WTSApplicationName: "ApplicationName",
    WTSWorkingDirectory: "WorkingDirectory",
    WTSOEMId: "OEMId",
    WTSSessionId: "SessionId",
    WTSUserName: "UserName",
    WTSWinStationName: "WinStationName",
    WTSDomainName: "DomainName",
    WTSConnectState: "ConnectState",
    WTSClientBuildNumber: "ClientBuildNumber",
    WTSClientName: "ClientName",
    WTSClientDirectory: "ClientDirectory",
    WTSClientProductId: "ClientProductId",
    WTSClientHardwareId: "ClientHardwareId",
    WTSClientAddress: "ClientAddress",
    WTSClientDisplay: "ClientDisplay",
    WTSClientProtocolType: "ClientProtocolType",
    WTSIdleTime: "IdleTime",
    WTSLogonTime: "LogonTime",
    WTSIncomingBytes: "IncomingBytes",
    WTSOutgoingBytes: "OutgoingBytes",
    WTSIncomingFrames: "IncomingFrames",
    WTSOutgoingFrames: "OutgoingFrames",
    WTSClientInfo: "ClientInfo",
    WTSSessionInfo: "SessionInfo",
    WTSSessionInfoEx: "SessionInfoEx",
    WTSConfigInfo: "ConfigInfo",
    WTSValidationInfo: "ValidationInfo",
    WTSSessionAddressV4: "SessionAddressV4",
    WTSIsRemoteSession: "IsRemoteSession",
}
WTSRegisterSessionNotification = wtsapi32.WTSRegisterSessionNotification
WTSRegisterSessionNotification.restype = BOOL
WTSRegisterSessionNotification.argtypes = [HANDLE, DWORD]
WTSUnRegisterSessionNotification = wtsapi32.WTSUnRegisterSessionNotification
WTSUnRegisterSessionNotification.restype = BOOL
WTSUnRegisterSessionNotification.argtypes = [HANDLE]
WTSOpenServerA = wtsapi32.WTSOpenServerA
WTSOpenServerA.restype = HANDLE
WTSOpenServerA.argtypes = [LPSTR]
WTSCloseServer = wtsapi32.WTSCloseServer
WTSCloseServer.restype = BOOL
WTSCloseServer.argtypes = [HANDLE]
WTSQuerySessionInformationA = wtsapi32.WTSQuerySessionInformationA
WTSQuerySessionInformationA.restype = BOOL
WTSQuerySessionInformationA.argtypes = [HANDLE, DWORD, c_int, POINTER(LPSTR), POINTER(DWORD)]
WTSFreeMemory = wtsapi32.WTSFreeMemory
WTSFreeMemory.restype = BOOL
# WTSFreeMemory.argtypes = [PVOID]
WTSEnumerateSessionsA = wtsapi32.WTSEnumerateSessionsA
WTSEnumerateSessionsA.restype = BOOL
WTSEnumerateSessionsA.argtypes = [HANDLE, DWORD, DWORD, PPWTS_SESSION_INFOA, PDWORD]
WTSEnumerateProcessesExA = wtsapi32.WTSEnumerateProcessesExA
WTSEnumerateProcessesExA.restype = BOOL
WTSEnumerateProcessesExA.argtypes = [HANDLE, PDWORD, DWORD, POINTER(LPSTR), PDWORD]

WTSDisconnectSession = wtsapi32.WTSDisconnectSession
WTSDisconnectSession.restype = BOOL
WTSDisconnectSession.argtypes = [HANDLE, DWORD, BOOL]
WTSLogoffSession = wtsapi32.WTSLogoffSession
WTSLogoffSession.restype = BOOL
WTSLogoffSession.argtypes = [HANDLE, DWORD, BOOL]
WTSSendMessageA = wtsapi32.WTSSendMessageA
WTSSendMessageA.restype = BOOL
WTSSendMessageA.argtypes = [HANDLE, DWORD, LPSTR, DWORD, LPSTR, DWORD, DWORD, DWORD, POINTER(DWORD), BOOL]
WTSShutdownSystem = wtsapi32.WTSShutdownSystem
WTSShutdownSystem.restype = BOOL
WTSShutdownSystem.argtypes = [HANDLE, DWORD]
WTSTerminateProcess = wtsapi32.WTSTerminateProcess
WTSTerminateProcess.restype = BOOL
WTSTerminateProcess.argtypes = [HANDLE, DWORD, DWORD]


# WTSVirtualChannelOpen
# WTSWaitSystemEvent


def get_session_info(session) -> dict[str, Any]:
    info: dict[str, Any] = {
        "StationName": session.pWinStationName.decode("latin1"),
        "State": CONNECT_STATE.get(session.State, str(session.State)),
    }
    csid = session.SessionId
    buf = LPSTR()
    size = DWORD()
    for q in (WTSInitialProgram, WTSApplicationName, WTSWorkingDirectory,
              WTSUserName, WTSDomainName,
              WTSClientName, WTSClientDirectory,
              ):
        if WTSQuerySessionInformationA(WTS_CURRENT_SERVER_HANDLE, csid, q, byref(buf), byref(size)):
            if buf.value:
                try:
                    info[WTS_INFO_CLASS.get(q, str(q))] = buf.value.decode("latin1")
                except UnicodeDecodeError:
                    pass
    if WTSQuerySessionInformationA(WTS_CURRENT_SERVER_HANDLE, csid, WTSClientDisplay, byref(buf), byref(size)):
        if size.value >= sizeof(WTS_CLIENT_DISPLAY):
            pdisplay = cast(buf, POINTER(WTS_CLIENT_DISPLAY))
            display = pdisplay[0]
            if display.HorizontalResolution > 0 and display.VerticalResolution > 0 and display.ColorDepth > 0:
                info["Display"] = {
                    "Width": display.HorizontalResolution,
                    "Height": display.VerticalResolution,
                    "Depth": display.ColorDepth,
                }
    # if WTSQuerySessionInformationA(WTS_CURRENT_SERVER_HANDLE, csid, WTSConnectState, byref(buf), byref(size)):
    #    if size.value==4:
    #        state = cast(buf, POINTER(DWORD)).contents.value
    #        info["ConnectState"] = CONNECT_STATE.get(state, state)
    if WTSQuerySessionInformationA(WTS_CURRENT_SERVER_HANDLE, csid, WTSClientProtocolType, byref(buf), byref(size)):
        if size.value == 2:
            ptype = cast(buf, POINTER(WORD)).contents.value
            info["Type"] = {0: "console", 1: "legacy", 2: "RDP"}.get(ptype, ptype)
    return info


def get_sessions() -> dict[int, dict[str, Any]]:
    cur = LPSTR(WTS_CURRENT_SERVER_HANDLE)
    h = WTSOpenServerA(cur)
    if not h:
        return {}
    session_info = PWTS_SESSION_INFOA()
    count = DWORD()
    sessions: dict[int, dict[str, Any]] = {}
    if WTSEnumerateSessionsA(h, 0, 1, byref(session_info), byref(count)):
        for i in range(count.value):
            session = session_info[i]
            sessions[int(session.SessionId)] = get_session_info(session)
        WTSFreeMemory(session_info)
    WTSCloseServer(h)
    return sessions


def find_session(username: str, with_display=True) -> dict:
    if username:
        for sid, info in get_sessions().items():
            if with_display and not info.get("Display"):
                continue
            if info.get("UserName", "").lower() == username.lower():
                info["SessionID"] = sid
                return info
    return {}


def main():
    import sys
    from xpra.platform.win32.common import WTSGetActiveConsoleSessionId
    csid = WTSGetActiveConsoleSessionId()
    print("WTSGetActiveConsoleSessionId()=%s" % csid)
    print_nested_dict(get_sessions())
    if len(sys.argv) > 1:
        for x in sys.argv[1:]:
            print("find_session(%s)=%s" % (x, find_session(x)))


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=wrong-import-position

import sys
import os.path
import subprocess
from collections.abc import Iterable, Sequence

from xpra.util.child_reaper import getChildReaper
from xpra.gtk.signals import register_os_signals
from xpra.gtk.window import add_close_accel
from xpra.gtk.widget import imagebutton, label
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.dialogs.util import hb_button
from xpra.scripts.main import run_example
from xpra.platform.paths import get_xpra_command
from xpra.os_util import WIN32, gi_import
from xpra.util.env import IgnoreWarningsContext
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gio = gi_import("Gio")

log = Logger("client", "util")


def exec_command(cmd: Sequence[str]) -> subprocess.Popen:
    env = os.environ.copy()
    env["XPRA_WAIT_FOR_INPUT"] = "0"
    creationflags = 0
    if WIN32:
        creationflags = subprocess.CREATE_NO_WINDOW  # @UndefinedVariable
    proc = subprocess.Popen(cmd, env=env, creationflags=creationflags)
    log("exec_command(%s)=%s", cmd, proc)
    return proc


TITLE = "Xpra Toolbox"

BUTTON_GROUPS: dict[str, Iterable[tuple[str, str, str]]] = {
    "Colors": (
        ("Squares", "Shows RGB+Grey squares in a window", "colors-plain"),
        ("Animated", "Shows RGB+Grey squares animated", "colors"),
        ("Bit Depth", "Shows color gradients and visualize bit depth clipping", "colors-gradient"),
    ),
    "Transparency and Rendering": (
        ("Circle", "Shows a semi-opaque circle in a transparent window", "transparent-window"),
        ("RGB Squares", "RGB+Black shaded squares in a transparent window", "transparent-colors"),
        ("OpenGL", "OpenGL window - transparent on some platforms", "opengl"),
    ),
    "Widgets": (
        ("Text Entry", "Simple text entry widget", "text-entry"),
        ("File Selector", "Open the file selector widget", "file-chooser"),
        ("Header Bar", "Window with a custom header bar", "header-bar"),
    ),
    "Events": (
        ("Grabs", "Test keyboard and pointer grabs", "grabs"),
        ("Clicks", "Double and triple click events", "clicks"),
        ("Focus", "Shows window focus events", "window-focus"),
    ),
    "Windows": (
        ("States", "Toggle various window attributes", "window-states"),
        ("Title", "Update the window title", "window-title"),
        ("Opacity", "Change window opacity", "window-opacity"),
        ("Transient", "Show transient windows", "window-transient"),
        ("Override Redirect", "Shows an override redirect window", "window-overrideredirect"),
    ),
    "Geometry": (
        ("Size constraints", "Specify window geometry size constraints", "window-geometry-hints"),
        ("Move-Resize", "Initiate move resize from application", "initiate-moveresize"),
    ),
    "Keyboard and Clipboard": (
        ("Keyboard", "Keyboard event viewer", "view-keyboard"),
        ("Clipboard", "Clipboard event viewer", "view-clipboard"),
    ),
    "Misc": (
        ("Tray", "Show a system tray icon", "tray"),
        ("Font Rendering", "Render characters with and without anti-aliasing", "fontrendering"),
        ("Bell", "Test system bell", "bell"),
        ("Cursors", "Show named cursors", "cursors"),
    ),
}


class ToolboxGUI(Gtk.Window):

    def __init__(self, title=TITLE):
        self.exit_code = 0
        self.start_session = None
        super().__init__()
        self.set_title(title)
        self.set_border_width(10)
        self.set_resizable(True)
        self.set_decorated(True)
        self.set_position(Gtk.WindowPosition.CENTER)
        with IgnoreWarningsContext():
            self.set_wmclass("xpra-toolbox", "Xpra-Toolbox")

        hb = Gtk.HeaderBar()
        hb.set_show_close_button(True)
        hb.props.title = title
        hb.add(hb_button("About", "help-about", self.show_about))
        hb.show_all()
        self.set_titlebar(hb)

        icon_name = "applications-utilities"
        self.set_icon_name(icon_name)
        icon_theme = Gtk.IconTheme.get_default()
        try:
            pixbuf = icon_theme.load_icon(icon_name, 96, 0)
        except Exception:
            pixbuf = get_icon_pixbuf("xpra")
        if pixbuf:
            self.set_icon(pixbuf)
        add_close_accel(self, self.quit)
        self.connect("delete_event", self.quit)

        self.vbox = Gtk.VBox(homogeneous=False, spacing=10)
        self.add(self.vbox)

        def addhbox(blabel: str, buttons: Iterable[tuple[str, str, str]]) -> None:
            self.vbox.add(label(blabel, font="sans 14"))
            hbox = Gtk.HBox(homogeneous=False, spacing=10)
            self.vbox.add(hbox)
            for button in buttons:
                if button:
                    hbox.add(self.button(*button))

        for category, button_defs in BUTTON_GROUPS.items():
            addhbox(f"{category}:", button_defs)
        self.vbox.show_all()

    @staticmethod
    def button(label_str, tooltip, example="", enabled=True) -> Gtk.Button:
        log(f"{label_str} : {example}")

        def cb(_btn) -> None:
            cmd = get_xpra_command() + ["example", example]
            proc = exec_command(cmd)
            getChildReaper().add_process(proc, label_str, cmd, ignore=True, forget=True)

        ib = imagebutton(label_str, None,
                         tooltip, clicked_callback=cb,
                         icon_size=48)
        ib.set_sensitive(enabled)
        return ib

    def quit(self, *args) -> None:
        log("quit%s", args)
        Gtk.main_quit()

    def app_signal(self, signum) -> None:
        self.exit_code = 128 + signum
        log("app_signal(%s) exit_code=%i", signum, self.exit_code)
        self.quit()

    def show_about(self, *_args) -> None:
        from xpra.gtk.dialogs.about import about
        about(parent=self)


def main(args) -> int:
    if len(args) == 1:
        # find a matching script or label:
        match = args[0].lower()
        return run_example(match)

    from xpra.platform import program_context
    from xpra.log import enable_color
    from xpra.platform.gui import init, ready, set_default_icon
    with program_context("Xpra-Toolbox", TITLE):
        enable_color()

        set_default_icon("toolbox.png")
        init()

        gui = ToolboxGUI()
        register_os_signals(gui.app_signal, TITLE)
        ready()
        gui.show()
        Gtk.main()
        log("do_main() gui.exit_code=%i", gui.exit_code)
        return gui.exit_code


if __name__ == "__main__":
    r = main(sys.argv[1:])
    sys.exit(r)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from time import sleep
from typing import Any
from collections.abc import Sequence

from xpra.server.proxy.server import ProxyServer as _ProxyServer, get_proxy_env
from xpra.platform.paths import get_app_dir
from xpra.common import SocketState
from xpra.util.env import envbool
from xpra.util.io import pollwait, which
from xpra.log import Logger

log = Logger("proxy")

PAEXEC = envbool("XPRA_PAEXEC", True)
SYSTEM_SESSION = envbool("XPRA_SYSTEM_SESSION", True)
TOKEN = envbool("XPRA_SYSTEM_TOKEN", False)


def exec_command(username: str, password: str, args: Sequence[str], exe: str, cwd: str, env: dict[str, str]):
    log("exec_command%s", (username, args, exe, cwd, env))
    # pylint: disable=import-outside-toplevel
    from xpra.platform.win32.lsa_logon_lib import logon_msv1, logon_msv1_s4u
    if TOKEN:
        logon_info = logon_msv1_s4u(username)
    else:
        logon_info = logon_msv1(username, password)
    log("logon(..)=%s", logon_info)
    from xpra.platform.win32.create_process_lib import (
        Popen,
        CREATIONINFO, CREATION_TYPE_LOGON,
        STARTF_USESHOWWINDOW,
        LOGON_WITH_PROFILE, CREATE_NEW_PROCESS_GROUP, STARTUPINFO,
    )
    creation_info = CREATIONINFO()
    creation_info.lpApplicationName = "Xpra"
    creation_info.lpUsername = username
    creation_info.lpPassword = password
    creation_info.lpDomain = os.environ.get("USERDOMAIN", "WORKGROUP")
    #creation_info.dwCreationType = CREATION_TYPE_TOKEN
    creation_info.dwCreationType = CREATION_TYPE_LOGON
    creation_info.dwLogonFlags = LOGON_WITH_PROFILE
    creation_info.dwCreationFlags = CREATE_NEW_PROCESS_GROUP
    creation_info.hToken = logon_info.Token
    log("creation_info=%s", creation_info)
    startupinfo = STARTUPINFO()
    startupinfo.dwFlags = STARTF_USESHOWWINDOW
    startupinfo.wShowWindow = 0  # aka win32.con.SW_HIDE
    startupinfo.lpDesktop = "WinSta0\\Default"
    startupinfo.lpTitle = "Xpra-Shadow"

    from subprocess import PIPE
    proc = Popen(args, executable=exe,
                 stdout=PIPE, stderr=PIPE,
                 cwd=cwd, env=env,
                 startupinfo=startupinfo, creationinfo=creation_info)
    log("Popen(%s)=%s", args, proc)
    return proc


class ProxyServer(_ProxyServer):

    def start_new_session(self, username: str, password: str, uid: int, gid: int,
                          sess_options: dict, displays=()) -> tuple[Any, str, str]:
        log("start_new_session%s", (username, "..", uid, gid, sess_options, displays))
        return self.start_win32_shadow(username, password, sess_options)

    def start_win32_shadow(self, username: str, password: str, sess_options: dict) -> tuple[Any, str, str]:
        log("start_win32_shadow%s", (username, "..", sess_options))
        app_dir = get_app_dir()
        shadow_command = app_dir + "\\Xpra-Shadow.exe"
        paexec = app_dir + "\\paexec.exe"
        named_pipe = username.replace(" ", "_")

        # pylint: disable=import-outside-toplevel
        from xpra.platform.win32.wtsapi import find_session
        session_info = find_session(username)
        if not session_info:
            # first, Logon:
            with log.trap_error(f"Error: failed to logon as {username!r}"):
                from xpra.platform.win32.desktoplogon_lib import logon
                r = logon(username, password)
            #if r:
            #    raise RuntimeError(f"desktop logon has failed and returned {r}")
        # hwinstaold = set_window_station("winsta0")
        wrap = []

        # use paexec to access the GUI session:
        if PAEXEC and os.path.exists(paexec) and os.path.isfile(paexec):
            # find the session-id to shadow:
            if not session_info:
                session_info = find_session(username)
            if session_info:
                log(f"found session {session_info} for {username!r}")
                wrap = [
                    "paexec.exe",
                    "-i", str(session_info["SessionID"]),
                ]
            elif username.lower() == "administrator" and SYSTEM_SESSION:
                log("using system session")
                wrap = [
                    "paexec.exe", "-x",
                ]
            else:
                log.warn("Warning: session not found for username '%s'", username)
        else:
            log(f"{PAEXEC=}, {paexec=!r}")
            log.warn("Warning: starting without paexec, expect a black screen")

        cmd = wrap + [
            shadow_command,
            f"--bind={named_pipe}",
            # "--tray=no",
        ]
        # unless explicitly stated otherwise, exit with client:
        if sess_options.get("exit-with-client", None) is not False:
            cmd.append("--exit-with-client=yes")
        from xpra.log import debug_enabled_categories
        if debug_enabled_categories:
            cmd += ["-d", ",".join(tuple(debug_enabled_categories))]
        env = get_proxy_env()
        env["XPRA_REDIRECT_OUTPUT"] = "1"
        exe = which(cmd[0])
        # env["XPRA_LOG_FILENAME"] = "Shadow-Instance.log"
        if username.lower() == "administrator" and SYSTEM_SESSION:
            # don't login
            from subprocess import Popen
            proc = Popen(cmd, executable=exe)
        else:
            proc = exec_command(username, password, cmd, exe, app_dir, env)

        from xpra.platform.win32.dotxpra import DotXpra
        dotxpra = DotXpra()
        for t in range(10):
            r = pollwait(proc, 1)
            if r is not None:
                log("pollwait=%s", r)
                try:
                    log("stdout=%s", proc.stdout.read())
                    log("stderr=%s", proc.stderr.read())
                except (OSError, AttributeError):
                    log("failed to read stdout / stderr of subprocess", exc_info=True)
                if r != 0:
                    raise RuntimeError(f"shadow subprocess failed with exit code {r}")
                raise RuntimeError("shadow subprocess has already terminated")
            if t >= 4:
                state = dotxpra.get_display_state(named_pipe)
                log("get_display_state(%s)=%s", named_pipe, state)
                if state == SocketState.LIVE:
                    # TODO: test the named pipe
                    sleep(2)
                    break
        self.child_reaper.add_process(proc, f"server-{username}", "xpra shadow", True, True)
        return proc, f"named-pipe://{named_pipe}", named_pipe



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os

from xpra.challenge.handler import AuthenticationHandler
from xpra.util.str_fn import csv
from xpra.os_util import WIN32
from xpra.log import Logger

log = Logger("auth")


def log_kerberos_exception(e):
    try:
        for x in e.args:
            if isinstance(x, (list, tuple)):
                try:
                    log.error(" %s", csv(x))
                    continue
                except Exception:
                    pass
            log.error(" %s", x)
    except Exception:
        log.estr(e)


class Handler(AuthenticationHandler):

    def __init__(self, **kwargs):
        services_str = kwargs.pop("kerberos-services", "") or os.environ.get("XPRA_KERBEROS_SERVICES", "") or "*"
        self.services = services_str.split(",")

    def __repr__(self):
        return "kerberos"

    def get_digest(self) -> str:
        return "kerberos"

    def handle(self, challenge: bytes, digest: str, prompt: str):  # pylint: disable=unused-argument
        if not digest.startswith("kerberos:"):
            log("%s is not a kerberos challenge", digest)
            # not a kerberos challenge
            return None
        try:
            # pylint: disable=import-outside-toplevel
            if WIN32:
                import winkerberos as kerberos
            else:
                import kerberos
        except ImportError as e:
            log.warn("Warning: cannot use kerberos authentication handler")
            log.warn(" %s", e)
            return False
        service = digest.split(":", 1)[1]
        if service not in self.services and "*" not in self.services:
            log.warn("Warning: invalid kerberos request for service '%s'", service)
            log.warn(" services supported: %s", csv(self.services))
            return None
        log("kerberos service=%s", service)
        try:
            r, ctx = kerberos.authGSSClientInit(service)  # @UndefinedVariable
            if r != 1:
                log("kerberos.authGSSClientInit failed and returned %s", r)
                return None
        except Exception as e:
            log("kerberos.authGSSClientInit(%s)", service, exc_info=True)
            log.error("Error: cannot initialize kerberos client:")
            log_kerberos_exception(e)
            return None
        try:
            kerberos.authGSSClientStep(ctx, "")  # @UndefinedVariable
        except Exception as e:
            log("kerberos.authGSSClientStep", exc_info=True)
            log.error("Error: kerberos client authentication failure:")
            log_kerberos_exception(e)
            return None
        token = kerberos.authGSSClientResponse(ctx)  # @UndefinedVariable
        log("kerberos token=%s", token)
        return token



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# This file is part of Xpra.
# Copyright (C) 2020 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# pylint: disable=line-too-long

import unittest

from xpra.net.protocol.header import (
    unpack_header, pack_header,
    FLAGS_BENCODE, FLAGS_RENCODE, FLAGS_CIPHER, FLAGS_YAML,
    LZ4_FLAG, BROTLI_FLAG,
)


class TestPacketHeader(unittest.TestCase):

    def test_roundtrip(self):
        print("hello")
        for encode_flag in (FLAGS_BENCODE, FLAGS_RENCODE, FLAGS_YAML):
            for comp_flag in (0, LZ4_FLAG, BROTLI_FLAG):
                for cipher in (0, FLAGS_CIPHER):
                    for level in (0, 1, 10):
                        for index in (0, 1, 255):
                            for size in (0, 1, 2**8, 2**16, 2**24):
                                proto_flags = encode_flag | comp_flag | cipher
                                header = pack_header(proto_flags, level, index, size)
                                assert header
                                uproto_flags, ulevel, uindex, usize = unpack_header(header)[1:]
                                assert uproto_flags==proto_flags
                                assert ulevel==level
                                assert uindex==index
                                assert usize==size
                                for i in range(0, len(header)-1):
                                    try:
                                        unpack_header(header[:i])
                                    except Exception:
                                        pass
                                    else:
                                        raise Exception("header unpacking should have failed for size %i" % i)


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Sequence, Any

from xpra.client.base.command import HelloRequestClient
from xpra.codecs.pillow.decoder import get_encodings, decompress
from xpra.exit_codes import ExitCode
from xpra.net.common import PacketType
from xpra.util.io import load_binary_file
from xpra.util.objects import typedict
from xpra.log import Logger

log = Logger("client", "encoding")

EXT_ALIASES = {"jpg": "jpeg"}


def get_client_base_classes() -> tuple[type, ...]:
    from xpra.client.base import features

    classes: list[type] = []
    # Warning: MmapClient must come first,
    # so it is initialized by the time HelloRequestClient creates the hello packet
    if features.mmap:
        from xpra.client.mixins.mmap import MmapClient
        classes.append(MmapClient)
    classes.append(HelloRequestClient)
    return tuple(classes)


CLIENT_BASES = get_client_base_classes()
ClientBaseClass = type('ClientBaseClass', CLIENT_BASES, {})


class EncodeClient(ClientBaseClass):
    """
    Sends the file(s) to the server for encoding,
    saves the result in the current working directory
    this requires a server version 6.3 or later
    """

    def __init__(self, options, filenames: Sequence[str]):
        if not filenames:
            raise ValueError("please specify some filenames to encode")
        for cc in CLIENT_BASES:
            args = (options, ) if cc == HelloRequestClient else ()
            cc.__init__(self, *args)
        self.filenames = list(filenames)
        self.add_packets("encode-response", "encodings")
        self.decompress = decompress
        self.encoding_options = {}
        self.encodings = get_encodings()

    def init(self, opts) -> None:
        if opts.mmap.lower() == "auto":
            opts.mmap = "yes"
        for cc in CLIENT_BASES:
            cc.init(self, opts)
        if opts.encoding and opts.encoding not in self.encodings:
            self.encodings = tuple(list(self.encodings) + [opts.encoding])
        self.encoding_options = {
            "options": self.encodings,
            "core": self.encodings,
            "setting": opts.encoding,
        }
        for attr, value in {
            "quality": opts.quality,
            "min-quality": opts.min_quality,
            "speed": opts.speed,
            "min-speed": opts.min_speed,
        }.items():
            if value > 0:
                self.encoding_options[attr] = value

    def setup_connection(self, conn) -> None:
        for cc in CLIENT_BASES:
            cc.setup_connection(self, conn)

    def server_connection_established(self, c: typedict) -> bool:
        for cc in CLIENT_BASES:
            if not cc.parse_server_capabilities(self, c):
                return False
        # this will call do_command()
        return super().server_connection_established(c)

    def client_type(self) -> str:
        return "encoder"

    def _process_encodings(self, packet: PacketType) -> None:
        encodings = typedict(packet[1]).dictget("encodings", {}).get("core", ())
        common = tuple(set(self.encodings) & set(encodings))
        log("server encodings=%s, common=%s", encodings, common)

    def _process_encode_response(self, packet: PacketType) -> None:
        encoding, data, options, width, height, bpp, stride, metadata = packet[1:9]
        log("encode-response: %8s %6i bytes, %5ix%-5i %ibits, stride=%i, options=%s, metadata=%s",
            encoding, len(data), width, height, bpp, stride, options, metadata)
        filename = typedict(metadata).strget("filename")
        if not filename:
            log.error("Error: 'filename' is missing from the metadata")
            self.quit(ExitCode.NO_DATA)
            return
        save_as = os.path.splitext(os.path.basename(filename))[0] + f".{encoding}"
        with open(save_as, "wb") as f:
            f.write(data)
        log.info(f"saved %i bytes to {save_as!r}", len(data))
        if not self.filenames:
            self.quit(ExitCode.OK)
            return
        self.send_encode()

    def hello_request(self) -> dict[str, Any]:
        hello = {}
        for cc in CLIENT_BASES:
            hello.update(cc.get_caps(self))
        hello = {
            "request": "encode",
            "ui_client": True,
            "encoding": self.encoding_options,
        }
        log(f"{hello=}")
        return hello

    def do_command(self, caps: typedict) -> None:
        log(f"{caps=}")
        self._protocol.large_packets.append("encode")
        self.send_encode()

    def send_encode(self):
        filename = self.filenames.pop(0)
        log(f"send_encode() {filename=!r}")
        ext = filename.split(".")[-1]
        ext = EXT_ALIASES.get(ext, ext)
        if ext not in self.encodings:
            log.warn(f"Warning: {ext!r} format is not supported")
            log.warn(" use %s", " or ".join(self.encodings))
            self.quit(ExitCode.UNSUPPORTED)
            return
        img_data = load_binary_file(filename)
        options = typedict()
        rgb_format, raw_data, width, height, rowstride = self.decompress(ext, img_data, options)
        # video encoders prefer BGRA pixel order:
        if rgb_format == "RGBA":
            from xpra.codecs.argb.argb import rgba_to_bgra
            raw_data = rgba_to_bgra(raw_data)
            rgb_format = "BGRX"
        elif rgb_format == "RGB":
            from xpra.codecs.argb.argb import rgb_to_bgrx
            raw_data = rgb_to_bgrx(raw_data)
            rgb_format = "BGRX"
            rowstride = rowstride * 4 // 3
        encoding = "rgb"
        encode_options = {}
        if self.compression_level > 0:
            from xpra.net.lz4.lz4 import compress
            data = compress(raw_data)
            encode_options["lz4"] = 1
            log("lz4 compressed from %i bytes down to %i", len(raw_data), len(data))
        else:
            log("sending uncompressed")
            data = raw_data

        mmap_write_area = getattr(self, "mmap_write_area", None)
        if mmap_write_area and mmap_write_area.enabled:
            mmap_data = mmap_write_area.write_data(data)
            log("mmap_write_area=%s, mmap_data=%s", mmap_write_area.get_info(), mmap_data)
            encoding = "mmap"
            data = b""
            encode_options["chunks"] = mmap_data

        metadata = {"filename": filename}
        self.send("encode", encoding, rgb_format, data, width, height, rowstride, encode_options, metadata)



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from queue import SimpleQueue
from typing import Union, Any
from collections.abc import Callable

from aioquic import __version__ as aioquic_version
from aioquic.h0.connection import H0Connection
from aioquic.h3.connection import H3Connection
from aioquic.h3.events import DataReceived, DatagramReceived, H3Event
from aioquic.quic.packet import QuicErrorCode

from xpra.net.quic.asyncio_thread import get_threaded_loop
from xpra.net.bytestreams import Connection
from xpra.net.quic.common import binary_headers, override_aioquic_logger
from xpra.util.str_fn import Ellipsizer, memoryview_to_bytes
from xpra.util.version import parse_version, vtrim
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("quic")

HttpConnection = Union[H0Connection, H3Connection]

DATAGRAM_PACKET_TYPES = tuple(x.strip() for x in os.environ.get(
    "XPRA_QUIC_DATAGRAM_PACKET_TYPES",
    ""
).split(",") if x.strip())

if envbool("XPRA_QUIC_LOGGER", True):
    override_aioquic_logger()


aioquic_version_info = parse_version(aioquic_version)


class XpraQuicConnection(Connection):
    def __init__(self, connection: HttpConnection, stream_id: int, transmit: Callable[[], None],
                 host: str, port: int, socktype="wss", info=None, options=None) -> None:
        Connection.__init__(self, (host, port), socktype, info=info, options=options)
        self.socktype_wrapped = "quic"
        self.connection: HttpConnection = connection
        self.read_queue: SimpleQueue[bytes] = SimpleQueue()
        self.stream_id: int = stream_id
        self.transmit: Callable[[], None] = transmit
        self.accepted: bool = False
        self.closed: bool = False

    def __repr__(self):
        return f"XpraQuicConnection<{self.socktype}:{self.stream_id}>"

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        qinfo = {
            "read-queue": self.read_queue.qsize(),
            "stream-id": self.stream_id,
            "accepted": self.accepted,
            "closed": self.closed,
            "aioquic": vtrim(aioquic_version_info),
        }
        quic = getattr(self.connection, "_quic", None)
        if quic:
            config = quic.configuration
            qinfo |= {
                "alpn-protocols": config.alpn_protocols,
                "idle-timeout": config.idle_timeout,
                "client": config.is_client,
                "max-data": config.max_data,
                "max-stream-data": config.max_stream_data,
                "server-name": config.server_name or "",
            }
        info["quic"] = qinfo
        return info

    def http_event_received(self, event: H3Event) -> None:
        log("quic:http_event_received(%s)", Ellipsizer(event))
        if self.closed:
            return
        if isinstance(event, (DataReceived, DatagramReceived)):
            self.read_queue.put(event.data)
        else:
            log.warn(f"Warning: unhandled websocket http event {event}")

    def close(self, code=QuicErrorCode.NO_ERROR, reason="closing") -> None:
        log(f"quic.close({code}, {reason})")
        if not self.closed:
            try:
                self.send_close(code, reason)
            finally:
                self.closed = True
        Connection.close(self)

    def send_close(self, code=QuicErrorCode.NO_ERROR, reason="") -> None:
        # we just close the quic connection here
        # subclasses may also override this method to send close messages specific to the protocol
        # ie: WebSocket and WebTransport 'close' frames
        quic = getattr(self.connection, "_quic", None)
        if not quic:
            return
        if aioquic_version_info >= (1, 2):
            # we can send the error code and message
            quic.close(code)
        else:
            quic.close()

    def send_headers(self, stream_id: int, headers: dict) -> None:
        self.connection.send_headers(
            stream_id=stream_id,
            headers=binary_headers(headers),
            end_stream=self.closed)

    def write(self, buf, packet_type: str = "") -> int:
        log("quic.write(%s, %r)", Ellipsizer(buf), packet_type)
        return self.stream_write(buf, packet_type)

    def stream_write(self, buf, packet_type: str) -> int:
        data = memoryview_to_bytes(buf)
        if not packet_type:
            log.warn(f"Warning: missing packet type for {data}")
        if packet_type in DATAGRAM_PACKET_TYPES:
            self.connection.send_datagram(self.stream_id, data=data)
            log(f"sending {packet_type!r} using datagram")
            return len(buf)
        stream_id = self.get_packet_stream_id(packet_type)
        log("quic.stream_write(%s, %s) using stream id %s", Ellipsizer(buf), packet_type, stream_id)

        def do_write() -> None:
            if self.closed:
                log(f"connection is already closed, packet {packet_type} dropped")
                return
            try:
                self.do_write(stream_id, data)
                self.transmit()
            except AssertionError:
                if self.closed:
                    log(f"connection is already closed, packet {packet_type} dropped")
                    return
                raise

        get_threaded_loop().call(do_write)
        return len(buf)

    def do_write(self, stream_id: int, data: bytes) -> None:
        self.connection.send_data(stream_id=stream_id, data=data, end_stream=self.closed)

    def get_packet_stream_id(self, packet_type: str) -> int:
        return self.stream_id

    def read(self, n: int) -> bytes:
        log("quic.read(%s)", n)
        return self.read_queue.get()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys

from xpra.auth.sqlauthbase import SQLAuthenticator, DatabaseUtilBase, run_dbutil
from xpra.auth.sys_auth_base import log


class Authenticator(SQLAuthenticator):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.uri = kwargs.get("uri")
        assert self.uri, "missing database uri"

    def db_cursor(self, *sqlargs):
        from sqlalchemy import create_engine  # pylint: disable=import-outside-toplevel
        db = create_engine(self.uri)
        cursor = db.cursor()
        cursor.execute(*sqlargs)
        # keep reference to db so that it doesn't get garbage collected just yet:
        cursor.db = db
        log("db_cursor(%s)=%s", sqlargs, cursor)
        return cursor

    def __repr__(self):
        return "sql"


class SQLDatabaseUtil(DatabaseUtilBase):

    def __init__(self, uri):
        super().__init__(uri)
        # from sqlalchemy import create_engine
        # db = create_engine(self.uri)
        self.param = os.environ.get("PARAMSTYLE", "%s")

    def exec_database_sql_script(self, cursor_cb, *sqlargs):
        from sqlalchemy import create_engine  # pylint: disable=import-outside-toplevel
        db = create_engine(self.uri)
        log("%s.execute%s", db, sqlargs)
        result = db.execute(*sqlargs)
        log("result=%s", result)
        if cursor_cb:
            cursor_cb(result)
        return result

    def get_authenticator_class(self):
        return Authenticator


def main(argv) -> int:
    return run_dbutil(SQLDatabaseUtil, "databaseURI", argv)


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.os_util import gi_import
from xpra.scripts.config import InitException
from xpra.net.common import PacketType
from xpra.x11.desktop.base import DesktopServerBase
from xpra.x11.desktop.monitor_model import MonitorDesktopModel
from xpra.server.mixins.window import WindowsMixin
from xpra.x11.vfb_util import parse_resolution
from xpra.x11.bindings.randr import RandRBindings
from xpra.gtk.error import xsync, xlog
from xpra.log import Logger

GObject = gi_import("GObject")
GLib = gi_import("GLib")

RandR = RandRBindings()

log = Logger("server")
metadatalog = Logger("x11", "metadata")
screenlog = Logger("screen")
iconlog = Logger("icon")

MIN_SIZE = 640, 350
MAX_SIZE = 8192, 8192


def get_screen_size() -> tuple[int, int]:
    with xsync:
        return RandR.get_screen_size()


class XpraMonitorServer(DesktopServerBase):
    """
        Virtualizes monitors
    """
    __gsignals__ = DesktopServerBase.__common_gsignals__

    def __init__(self):
        with xsync:
            if not RandR.is_dummy16():
                display = os.environ.get("DISPLAY", "")
                raise InitException(f"the vfb display {display!r} cannot virtualize monitors - dummy RandR 1.6 missing")
        super().__init__()
        self.session_type: str = "monitor"
        self.reconfigure_timer: int = 0
        self.reconfigure_locked: bool = False

    def init_randr(self) -> None:
        super().init_randr()
        from xpra.x11.vfb_util import set_initial_resolution
        screenlog(f"init_randr() randr={self.randr}, initial-resolutions={self.initial_resolutions}")
        if not RandR.has_randr() or not self.initial_resolutions:
            return
        res = self.initial_resolutions
        with xlog:
            set_initial_resolution(res, self.dpi or self.default_dpi)

    def get_server_mode(self) -> str:
        return "X11 monitor"

    def make_hello(self, source) -> dict[str, Any]:
        capabilities = super().make_hello(source)
        if "features" in source.wants:
            capabilities |= {
                "monitor": True,
                "multi-monitors": True,
                "monitors": self.get_monitor_config(),
                "monitors.min-size": MIN_SIZE,
                "monitors.max-size": MAX_SIZE,
            }
        return capabilities

    def configure_best_screen_size(self):
        sss = tuple(x for x in self._server_sources.values() if x.ui_client)
        log(f"configure_best_screen_size() sources={sss}")
        if len(sss) != 1:
            screenlog.info(f"screen used by {len(sss)} clients:")
            return get_screen_size()
        ss = sss[0]
        if not getattr(ss, "desktop_fullscreen", False):
            return get_screen_size()
        # try to match this client's layout:
        log("will try to mirror")
        # prevent this monitor layout change
        # from triggering a call to via reconfigure_monitors via reconfigure:
        try:
            self.reconfigure_locked = True
            mdef = self.mirror_client_monitor_layout()
            if mdef:
                self.setting_changed("monitors", mdef)
        except Exception:
            log("Warning: failed to mirror client monitor layout", exc_info=True)
            self.reconfigure_locked = False

        def unlock():
            self.reconfigure_locked = False

        GLib.timeout_add(1000, unlock)
        return get_screen_size()

    def load_existing_windows(self) -> None:
        with xlog:
            monitors = RandR.get_monitor_properties()
            for i, monitor in monitors.items():
                self.add_monitor_model(i + 1, monitor)
        # does not fire: (because of GTK?)
        # RandR.select_crtc_output_changes()
        # screen = gdk.Screen.get_default()
        # screen.connect("monitors-changed", self.monitors_changed)

    def do_x11_configure_event(self, event) -> None:
        # the root window changed,
        # check to see if a monitor has been modified
        # do this via a timer to avoid running multiple times
        # as we get multiple events for the same change
        log("do_x11_configure_event(%s)", event)
        if not self.reconfigure_timer:
            self.reconfigure_timer = GLib.timeout_add(50, self.reconfigure)

    def reconfigure(self) -> None:
        try:
            self.do_reconfigure()
        finally:
            self.reconfigure_timer = 0

    def do_reconfigure(self) -> None:
        # verify that our models are up-to-date,
        # we look for the `crtcs` because that's what tools like `xrandr` can modify easily
        #  ie: `xrandr --output DUMMY1 --mode 1024x768`
        mdefs = {}
        with xlog:
            info = RandR.get_all_screen_properties()
            crtcs = info.get("crtcs")
            outputs = info.get("outputs")
            monitors = info.get("monitors")

            def find_monitor(output_id):
                for minfo in monitors.values():
                    if output_id in minfo.get("outputs", ()):
                        return minfo
                return None

            screenlog("do_reconfigure() crtcs=%s", crtcs)
            screenlog("do_reconfigure() outputs=%s", outputs)
            screenlog("do_reconfigure() monitors=%s", monitors)
            if not crtcs or not outputs:
                return
            for i, crtc_info in crtcs.items():
                # find the monitor for this crtc:
                if crtc_info.get("noutput", 0) != 1:
                    screenlog(f"no outputs on crtc {i}")
                    continue
                output_id = crtc_info.get("outputs")[0]
                output_info = outputs.get(output_id)
                if not output_info:
                    screenlog(f"output {output_id} not found")
                    continue
                if output_info.get("connection") != "Connected":
                    screenlog(f"output {output_id} is not connected")
                    continue
                monitor_info = find_monitor(output_id)
                if not monitor_info:
                    screenlog(f"no monitor found for output id {output_id}")
                    return
                # get the geometry from the crtc:
                mdef = {k: v for k, v in crtc_info.items() if k in ("x", "y", "width", "height")}
                # add the millimeter dimensions from the output:
                mdef.update((k, v) for k, v in output_info.items() if k in ("mm-width", "mm-height"))
                # and some monitor attributes:
                mdef.update((k, v) for k, v in output_info.items() if k in ("primary", "automatic", "name"))
                mdefs[i] = mdef
                screenlog(f"do_reconfigure() {i}: {mdef}")
            if self.sync_monitors_to_models(mdefs) and not self.reconfigure_locked:
                self.reconfigure_monitors()
        self.refresh_all_windows()

    def sync_monitors_to_models(self, monitors) -> int:
        # now update the monitor models with this data:
        screenlog("sync_monitors_to_models(%s)", monitors)
        mods = 0
        for i, monitor in monitors.items():
            wid = i + 1
            model = self._id_to_window.get(wid)
            if not model:
                # found a new monitor!
                screenlog("found a new monitor: %s", monitor)
                self.add_monitor_model(wid, monitor)
                continue
            mdef = model.get_definition()
            diff = [k for k in ("x", "y", "width", "height") if monitor.get(k) != mdef.get(k)]
            screenlog("model %i geometry modified %s from %s to %s",
                      wid, diff, [mdef.get(k) for k in diff], [monitor.get(k) for k in diff])
            if diff:
                # re-initialize with new geometry:
                log("was %s, now %s", mdef, monitor)
                model.init(monitor)
                model.emit("resized")
                mods += 1
            if mdef.get("name", "") != monitor.get("name", ""):
                model.name = monitor.get("name")
                screenlog(f"monitor {i} name has changed to {model.name!r}")
                # name is used to generate the window "title":
                model.notify("title")
        return mods

    def add_monitor_model(self, wid: int, monitor) -> MonitorDesktopModel:
        model = MonitorDesktopModel(monitor)
        model.setup()
        screenlog("adding monitor model %s", model)
        super().do_add_new_window_common(wid, model)
        model.managed_connect("client-contents-changed", self._contents_changed)
        model.managed_connect("resized", self.monitor_resized)
        model.managed_connect("motion", self._motion_signaled)
        return model

    def monitor_resized(self, model) -> None:
        delta_x, delta_y = model.resize_delta
        wid = self._window_to_id[model]
        w, h = model.get_dimensions()
        screenlog("monitor_resized(%s) size=%s, delta=%s, wid=%i",
                  model, (w, h), (delta_x, delta_y), wid)
        for ss in self.window_sources():
            ss.resize_window(wid, model, w, h)
        if self.reconfigure_timer:
            # we're in the process of adjusting things
            return
        # we adjust the position of monitors after this one,
        # assuming that they are defined left to right!
        # first the models:
        self._adjust_monitors(wid, delta_x, delta_y)
        self.reconfigure_monitors()

    def reconfigure_monitors(self) -> None:
        # now we can do the virtual crtcs, outputs and monitors
        defs = self.get_monitor_config()
        screenlog("reconfigure_monitors() definitions=%s", defs)
        self.apply_monitor_config(defs)
        # and tell the client:
        self.setting_changed("monitors", defs)

    def validate_monitors(self) -> None:
        for model in self._id_to_window.values():
            x, y, width, height = model.get_geometry()
            if x + width >= MAX_SIZE[0] or y + height >= MAX_SIZE[1]:
                new_x, new_y = 0, 0
                mdef = model.get_definition()
                mdef |= {
                    "x": new_x,
                    "y": new_y,
                }
                model.init(mdef)

    def _adjust_monitors(self, after_wid: int, delta_x: int, delta_y: int) -> None:
        models = {wid: model for wid, model in self._id_to_window.items() if wid > after_wid}
        screenlog("adjust_monitors(%i, %i, %i) models=%s", after_wid, delta_x, delta_y, models)
        if (delta_x == 0 and delta_y == 0) or not models:
            return
        for wid, model in models.items():
            self._adjust_monitor(model, delta_x, delta_y)

    def _adjust_monitor(self, model, delta_x: int, delta_y: int) -> None:
        screenlog("adjust_monitors(%s, %i, %i)", model, delta_x, delta_y)
        if delta_x == 0 and delta_y == 0:
            return
        x, y = model.get_geometry()[:2]
        new_x = max(0, x + delta_x)
        new_y = max(0, y + delta_y)
        if new_x != x or new_y != y:
            screenlog(f"adjusting monitor {model} from {x},{y} to {new_x},{new_y}")
            mdef = model.get_definition()
            mdef |= {
                "x": new_x,
                "y": new_y,
            }
            model.init(mdef)

    def get_monitor_config(self) -> dict[int, dict]:
        monitor_defs = {}
        for wid, model in self._id_to_window.items():
            monitor = model.get_definition()
            i = wid - 1
            monitor["index"] = i
            monitor_defs[i] = monitor
        return monitor_defs

    def apply_monitor_config(self, monitor_defs: dict) -> None:
        with xsync:
            RandR.set_crtc_config(monitor_defs)

    def remove_monitor(self, wid: int) -> None:
        model = self._id_to_window.get(wid)
        screenlog("removing monitor for wid %i : %s", wid, model)
        if not model:
            raise ValueError(f"monitor {wid} not found")
        if len(self._id_to_window) <= 1:
            raise RuntimeError("cannot remove the last monitor")
        delta_x = -model.get_definition().get("width", 0)
        delta_y = 0  # model.monitor.get("width", 0)
        model.unmanage()
        wid = self._remove_window(model)
        # adjust the position of the other monitors:
        self._adjust_monitors(wid, delta_x, delta_y)
        self.reconfigure_monitors()

    def add_monitor(self, width: int, height: int) -> None:
        count = len(self._id_to_window)
        if count >= 16:
            raise RuntimeError(f"already too many monitors: {count}")
        if not (isinstance(width, int) and isinstance(height, int)):
            raise ValueError(f"invalid dimension types: {width} ({type(width)}) and {height} ({type(height)})")
        if (width, height) < MIN_SIZE:
            raise ValueError(f"monitor size {width}x{height} is too small, minimum is {MIN_SIZE}")
        if (width, height) > MAX_SIZE:
            raise ValueError(f"monitor size {width}x{height} is too large, maximum is {MIN_SIZE}")

        # find the wid to use:
        # prefer just incrementing the wid, but we cannot go higher than 16

        def rightof(wid: int):
            mdef = self._id_to_window[wid].get_definition()
            x = mdef.get("x", 0) + mdef.get("width", 0)
            y = mdef.get("y", 0)  # +monitor.get("height", 0)
            return x, y

        wid = self._max_window_id
        x = y = 0
        if wid < 16:
            # since we're just appending,
            # just place to the right of the last monitor:
            last = max(self._id_to_window)
            x, y = rightof(last)
        else:
            # find a gap we can use in the window ids before 16:
            prev = None
            for wid in range(16):
                if wid not in self._id_to_window:
                    break
                prev = wid
            assert wid <= 16
            if prev:
                x, y = rightof(prev)
            self._adjust_monitors(wid - 1, width, 0)
        # ensure no monitors end up too far to the right or bottom:
        # (better have them overlap - though we could do something smarter here)
        self.validate_monitors()
        # now we can add our new monitor:
        xdpi = self.xdpi or self.dpi or 96
        ydpi = self.ydpi or self.dpi or 96
        wmm = round(width * 25.4 / xdpi)
        hmm = round(height * 25.4 / ydpi)
        index = wid - 1
        monitor = {
            "index": index,
            "name": f"VFB-{index}",
            "x": x,
            "y": y,
            "width": width,
            "height": height,
            "mm-width": wmm,
            "mm-height": hmm,
        }
        with xsync:
            model = self.add_monitor_model(wid, monitor)
        self.reconfigure_monitors()
        # send it to the clients:
        for ss in self._server_sources.values():
            if not isinstance(ss, WindowsMixin):
                continue
            self.send_new_desktop_model(model, ss)

    def _process_configure_monitor(self, _proto, packet: PacketType) -> None:
        action = str(packet[1])
        if action == "remove":
            identifier = str(packet[2])
            value = packet[3]
            if identifier == "wid":
                wid = int(value)
            elif identifier == "index":
                # index is zero-based
                wid = int(value) + 1
            else:
                raise ValueError(f"unsupported monitor identifier {identifier!r}")
            self.remove_monitor(wid)
        elif action == "add":
            resolution = packet[2]
            if isinstance(resolution, str):
                resolution = parse_resolution(resolution, self.refresh_rate)
            if not isinstance(resolution, (tuple, list)):
                raise ValueError(f"invalid resolution: {resolution!r} ({type(resolution)}")
            if len(resolution) not in (2, 3) or not all(isinstance(res, int) for res in resolution):
                raise ValueError(f"invalid resolution type: {resolution!r}")
            width, height = resolution[:2]
            self.add_monitor(width, height)
        else:
            raise ValueError(f"unsupported 'configure-monitor' action {action!r}")
        self.refresh_all_windows()

    def init_packet_handlers(self) -> None:
        super().init_packet_handlers()
        self.add_packets("configure-monitor", main_thread=True)


GObject.type_register(XpraMonitorServer)



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


# Sometimes we want to open the X11 display with XOpenDisplay,
# other times we may get an existing "Display" pointer from
# somewhere else, so we need this file to hide that...
# (we can't just pass pointers around easily with Python/Cython)

from libc.stdint cimport uintptr_t   # pylint: disable=syntax-error
from xpra.x11.bindings.xlib cimport Display

cdef Display *display = NULL
display_name = ""


cdef Display* get_display() noexcept:
    return display


def get_display_ptr() -> long:
    return int(<uintptr_t> display)


cdef int set_display(Display *d) except 1:
    global display
    if display!=NULL and d!=NULL and d!=display:
        raise RuntimeError("display is already set")
    display = d
    return 0


def clear_display() -> None:
    global display
    display = NULL


def get_display_name() -> str:
    return display_name


def set_display_name(name: str) -> None:
    global display_name
    display_name = name



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Sequence

from cairo import Format

from xpra.os_util import gi_import
from xpra.util.env import envbool
from xpra.util.objects import typedict
from xpra.client.gtk3.cairo_backing_base import CairoBackingBase, FORMATS
from xpra.gtk.cairo_image import make_image_surface, CAIRO_FORMATS
from xpra.log import Logger

log = Logger("paint", "cairo")

GLib = gi_import("GLib")
GdkPixbuf = gi_import("GdkPixbuf")

CAIRO_USE_PIXBUF = envbool("XPRA_CAIRO_USE_PIXBUF", False)


class CairoBacking(CairoBackingBase):
    """
    An area we draw onto with cairo
    This requires `cairo_bindings`.
    """

    RGB_MODES: Sequence[str] = ("BGRA", "BGRX", "RGBA", "RGBX", "BGR", "RGB", "r210", "BGR565")

    def __repr__(self):
        b = self._backing
        if b:
            binfo = "ImageSurface(%i, %i)" % (b.get_width(), b.get_height())
        else:
            binfo = "None"
        return "bindings.CairoBacking(%s : size=%s, render_size=%s)" % (binfo, self.size, self.render_size)

    def _do_paint_rgb(self, cairo_format, has_alpha: bool, img_data,
                      x: int, y: int, width: int, height: int, render_width: int, render_height: int,
                      rowstride: int, options: typedict) -> None:
        """ must be called from UI thread """
        log("cairo._do_paint_rgb%s make_image_surface=%s, use pixbuf=%s",
            (FORMATS.get(cairo_format, cairo_format), has_alpha, len(img_data),
             type(img_data), x, y, width, height, render_width, render_height,
             rowstride, options), make_image_surface, CAIRO_USE_PIXBUF)
        rgb_format = options.strget("rgb_format", "RGB")
        if not CAIRO_USE_PIXBUF:
            rgb_formats = CAIRO_FORMATS.get(cairo_format, ())
            if rgb_format in rgb_formats:
                img_surface = make_image_surface(cairo_format, rgb_format, img_data, width, height, rowstride)
                self.cairo_paint_surface(img_surface, x, y, render_width, render_height, options)
                img_surface.finish()
                return
            log("cannot set image surface data for cairo format %s and rgb_format %s (rgb formats supported: %s)",
                FORMATS.get(cairo_format, cairo_format), rgb_format, rgb_formats)

        if rgb_format in ("RGB", "RGBA", "RGBX"):
            data = GLib.Bytes(img_data)
            pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(data, GdkPixbuf.Colorspace.RGB,
                                                     has_alpha, 8, width, height, rowstride)
            if render_width != width or render_height != height:
                resample = options.strget("resample", "bilinear")
                if resample == "NEAREST":
                    interp_type = GdkPixbuf.InterpType.NEAREST
                elif resample in ("BICUBIC", "LANCZOS"):
                    interp_type = GdkPixbuf.InterpType.HYPER
                else:
                    interp_type = GdkPixbuf.InterpType.BILINEAR
                log(f"scaling using {resample!r} from {width}x{height} to {render_width}x{render_height}")
                pixbuf = pixbuf.scale_simple(render_width, render_height, interp_type)
            self.cairo_paint_pixbuf(pixbuf, x, y, options)
            return

        raise ValueError(f"failed to paint {cairo_format}")

    def update_fps_buffer(self, width: int, height: int, pixels) -> None:
        self.fps_image = make_image_surface(Format.ARGB32, "RGBA", pixels, width, height, width * 4)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2014 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
import sys
import time
from typing import Any
from collections.abc import Callable

from xpra.common import ScreenshotData, noop
from xpra.gtk.util import get_default_root_window
from xpra.gtk.window import add_close_accel
from xpra.gtk.info import get_display_info
from xpra.gtk.widget import scaled_image, label, choose_file
from xpra.gtk.pixbuf import get_icon_pixbuf
from xpra.gtk.versions import get_gtk_version_info
from xpra.platform.gui import force_focus
from xpra.os_util import gi_import
from xpra.exit_codes import ExitValue
from xpra.util.str_fn import nonl, repr_ellipsized, hexstr
from xpra.util.objects import typedict
from xpra.util.env import envint
from xpra.common import FULL_INFO
from xpra.log import Logger

log = Logger("util")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")

STEP_DELAY = envint("XPRA_BUG_REPORT_STEP_DELAY", 0)


def get_pillow_imagegrab_fn() -> Callable:
    try:
        from PIL import ImageGrab
        from io import BytesIO
    except ImportError as e:
        log("cannot use Pillow's ImageGrab: %s", e)
        return noop

    def pillow_imagegrab() -> ScreenshotData:
        img = ImageGrab.grab()
        out = BytesIO()
        img.save(out, format="PNG")
        pixels = out.getvalue()
        out.close()
        return img.width, img.height, "png", img.width * 3, pixels

    return pillow_imagegrab


class BugReport:

    def __init__(self):
        self.checkboxes = {}
        self.server_log = None
        self.show_about = True
        self.get_server_info: Callable | None = None
        self.opengl_info: dict = {}
        self.includes: dict = {}
        self.window: Gtk.Window | None = None
        self.description: Gtk.TextView | None = None
        self.toggles: tuple = ()

    def init(self, show_about: bool = True,
             get_server_info: Callable | None = None,
             opengl_info=None, includes=None):
        self.show_about = show_about
        self.get_server_info = get_server_info
        self.opengl_info = opengl_info
        self.includes = includes or {}
        self.setup_window()

    def setup_window(self) -> None:
        self.window = Gtk.Window()
        self.window.set_border_width(20)
        self.window.connect("delete-event", self.close)
        self.window.set_default_size(400, 300)
        self.window.set_title("Xpra Bug Report")

        icon_pixbuf = get_icon_pixbuf("bugs.png")
        if icon_pixbuf:
            self.window.set_icon(icon_pixbuf)
        self.window.set_position(Gtk.WindowPosition.CENTER)

        vbox = Gtk.VBox(homogeneous=False, spacing=0)
        vbox.set_spacing(15)

        # Title
        hbox = Gtk.HBox(homogeneous=False, spacing=0)
        icon_pixbuf = get_icon_pixbuf("xpra.png")
        if icon_pixbuf and self.show_about:
            from xpra.gtk.dialogs.about import about
            logo_button = Gtk.Button(label="")
            settings = logo_button.get_settings()
            settings.set_property('gtk-button-images', True)

            def show_about(*_args):
                about(parent=self.window)

            logo_button.connect("clicked", show_about)
            logo_button.set_tooltip_text("About")
            image = Gtk.Image()
            image.set_from_pixbuf(icon_pixbuf)
            logo_button.set_image(image)
            hbox.pack_start(logo_button, expand=False, fill=False)

        # the box containing all the input:
        ibox = Gtk.VBox(homogeneous=False, spacing=0)
        ibox.set_spacing(3)
        vbox.pack_start(ibox)

        # Description
        al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0)
        al.add(label("Please describe the problem:"))
        ibox.pack_start(al)
        self.description = Gtk.TextView()
        self.description.set_accepts_tab(True)
        self.description.set_justification(Gtk.Justification.LEFT)
        self.description.set_border_width(2)
        self.description.set_size_request(300, 80)
        ibox.pack_start(self.description, expand=False, fill=False)

        # Toggles:
        al = Gtk.Alignment(xalign=0, yalign=0.5, xscale=0.0, yscale=0)
        al.add(label("Include:"))
        ibox.pack_start(al)
        # generic toggles:
        from xpra.gtk.keymap import get_gtk_keymap
        from xpra.codecs.loader import codec_versions, load_codecs, show_codecs
        load_codecs()
        show_codecs()
        try:
            from xpra.audio.wrapper import query_audio

            def get_audio_info() -> typedict:
                return query_audio()
        except ImportError:
            get_audio_info = None

        def get_gl_info():
            if self.opengl_info:
                return self.opengl_info

        from xpra.net.net_util import get_info as get_net_info
        from xpra.platform.paths import get_info as get_path_info
        from xpra.platform.gui import get_info as get_gui_info
        from xpra.util.version import get_version_info, get_platform_info, get_host_info

        def get_sys_info() -> dict[str, Any]:
            from xpra.platform.info import get_user_info
            from xpra.scripts.config import read_xpra_defaults
            return {
                "argv": sys.argv,
                "path": sys.path,
                "exec_prefix": sys.exec_prefix,
                "executable": sys.executable,
                "version": get_version_info(),
                "platform": get_platform_info(),
                "host": get_host_info(FULL_INFO),
                "paths": get_path_info(),
                "gtk": get_gtk_version_info(),
                "gui": get_gui_info(),
                "display": get_display_info(),
                "user": get_user_info(),
                "env": os.environ,
                "config": read_xpra_defaults(),
            }

        get_screenshot: Callable = noop
        take_screenshot_fn: Callable = noop
        # screenshot: may have OS-specific code
        try:
            from xpra.platform.gui import take_screenshot
            take_screenshot_fn = take_screenshot
        except ImportError:
            log("failed to load platform specific screenshot code", exc_info=True)
        if take_screenshot_fn == noop:
            # try with Pillow:
            take_screenshot_fn = get_pillow_imagegrab_fn()
        if take_screenshot_fn == noop:
            # default: gtk screen capture
            try:
                from xpra.server.shadow.gtk_root_window_model import GTKImageCapture
                rwm = GTKImageCapture(get_default_root_window())
                take_screenshot_fn = rwm.take_screenshot
            except (ImportError, AttributeError):
                log.warn("Warning: failed to load gtk screenshot code", exc_info=True)
        log("take_screenshot_fn=%s", take_screenshot_fn)
        if take_screenshot_fn != noop:
            def _get_screenshot():
                # take_screenshot() returns: w, h, "png", rowstride, data
                return take_screenshot_fn()[4]

            get_screenshot = _get_screenshot

        def get_server_log():
            return self.server_log

        self.toggles = (
            ("system", "txt", "System", get_sys_info, True,
             "Xpra version, platform and host information - including hostname and account information"),
            ("server-log", "txt", "Server Log", get_server_log, bool(self.server_log),
             "Xpra version, platform and host information - including hostname and account information"),
            ("network", "txt", "Network", get_net_info, True,
             "Compression, packet encoding and encryption"),
            ("encoding", "txt", "Encodings", codec_versions, bool(codec_versions),
             "Picture encodings supported"),
            ("opengl", "txt", "OpenGL", get_gl_info, bool(self.opengl_info),
             "OpenGL driver and features"),
            ("audio", "txt", "Audio", get_audio_info, bool(get_audio_info),
             "Audio codecs and GStreamer version information"),
            ("keyboard", "txt", "Keyboard Mapping", get_gtk_keymap, True,
             "Keyboard layout and key mapping"),
            ("xpra-info", "txt", "Server Info", self.get_server_info, bool(self.get_server_info),
             "Full server information from 'xpra info'"),
            ("screenshot", "png", "Screenshot", get_screenshot, bool(get_screenshot),
             ""),
        )
        self.checkboxes = {}
        for name, _, title, value_cb, sensitive, tooltip in self.toggles:
            cb = Gtk.CheckButton(label=title + [" (not available)", ""][bool(value_cb)])
            cb.set_active(self.includes.get(name, True))
            cb.set_sensitive(sensitive)
            cb.set_tooltip_text(tooltip)
            ibox.pack_start(cb)
            self.checkboxes[name] = cb

        # Buttons:
        hbox = Gtk.HBox(homogeneous=False, spacing=20)
        vbox.pack_start(hbox)

        def btn(text, tooltip_text, callback, icon_name=None) -> Gtk.Button:
            b = Gtk.Button(label=text)
            b.set_tooltip_text(tooltip_text)
            b.connect("clicked", callback)
            if icon_name:
                icon = get_icon_pixbuf(icon_name)
                if icon:
                    b.set_image(scaled_image(icon, 24))
            hbox.pack_start(b)
            return b

        btn("Copy to clipboard", "Copy all data to clipboard", self.copy_clicked, "clipboard.png")
        btn("Save", "Save Bug Report", self.save_clicked, "download.png")
        btn("Cancel", "", self.close, "quit.png")

        def accel_close(*_args) -> None:
            self.close()

        add_close_accel(self.window, accel_close)
        vbox.show_all()
        self.window.vbox = vbox
        self.window.add(vbox)

    def set_server_log_data(self, filedata) -> None:
        self.server_log = filedata
        cb = self.checkboxes.get("server-log")
        log("set_server_log_data(%i bytes) cb=%s", len(filedata or b""), cb)
        if cb:
            cb.set_sensitive(bool(filedata))

    def show(self) -> None:
        log("show()")
        if not self.window:
            self.setup_window()
        force_focus()
        self.window.show_all()
        self.window.present()

    def hide(self) -> None:
        log("hide()")
        if self.window:
            self.window.hide()

    def close(self, *_args) -> bool:
        self.hide_window()
        return True

    def hide_window(self) -> None:
        log("hide_window()")
        if self.window:
            self.hide()
            self.window = None

    def destroy(self, *args) -> None:
        log("destroy%s", args)
        if self.window:
            self.window.destroy()
            self.window = None

    @staticmethod
    def run() -> ExitValue:
        log("run()")
        Gtk.main()
        log("run() Gtk.main done")
        return 0

    def quit(self, *args) -> None:
        log("quit%s", args)
        self.hide_window()
        Gtk.main_quit()

    def get_data(self) -> list[tuple[str, str, str, str]]:
        log("get_data() collecting bug report data")
        data = []
        tb = self.description.get_buffer()
        buf = tb.get_text(*tb.get_bounds(), include_hidden_chars=False)
        if buf:
            data.append(("Description", "", "txt", buf))
        for name, dtype, title, value_cb, _, tooltip in self.toggles:
            if not bool(value_cb):
                continue
            cb = self.checkboxes.get(name)
            assert cb is not None
            if not cb.get_active():
                continue
            log("%s is enabled (%s)", name, tooltip)
            # OK, the checkbox is selected, get the data
            value = value_cb
            if not isinstance(value_cb, dict):
                try:
                    value = value_cb()
                except TypeError:
                    log.error("Error collecting %s bug report data using %s", name, value_cb, exc_info=True)
                    value = str(value_cb)
                    dtype = "txt"
                except Exception as e:
                    log.error("Error collecting %s bug report data using %s", name, value_cb, exc_info=True)
                    value = e
                    dtype = "txt"
            if value is None:
                s = "not available"
            elif isinstance(value, dict):
                s = os.linesep.join("%s : %s" % (k.ljust(32), nonl(str(v))) for k, v in sorted(value.items()))
            elif isinstance(value, (list, tuple)):
                s = os.linesep.join(str(x) for x in value)
            else:
                s = value
            log("%s (%s) %s: %s", title, tooltip, dtype, repr_ellipsized(s))
            data.append((title, tooltip, dtype, s))
            time.sleep(STEP_DELAY)
        return data

    def copy_clicked(self, *_args) -> None:
        data = self.get_data()

        def cdata(v) -> str:
            if isinstance(v, bytes):
                return hexstr(v)
            return str(v)

        text = os.linesep.join("%s: %s%s%s%s" % (title, tooltip, os.linesep, cdata(v), os.linesep)
                               for (title, tooltip, dtype, v) in data if dtype == "txt")
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(text, len(text))
        log.info("%s characters copied to clipboard", len(text))

    def save_clicked(self, *_args) -> None:
        file_filter = Gtk.FileFilter()
        file_filter.set_name("ZIP")
        file_filter.add_pattern("*.zip")
        choose_file(self.window, "Save Bug Report Data", Gtk.FileChooserAction.SAVE, Gtk.STOCK_SAVE, self.do_save)

    def do_save(self, filename: str) -> None:
        log("do_save(%s)", filename)
        if not filename.lower().endswith(".zip"):
            filename = filename + ".zip"
        basenoext = os.path.splitext(os.path.basename(filename))[0]
        data = self.get_data()
        import zipfile
        zf = None
        try:
            zf = zipfile.ZipFile(filename, mode='w', compression=zipfile.ZIP_DEFLATED)
            for title, tooltip, dtype, s in data:
                cfile = os.path.join(basenoext, title.replace(" ", "_") + "." + dtype)
                # noinspection PyTypeChecker
                info = zipfile.ZipInfo(cfile, date_time=tuple(time.localtime(time.time())))
                info.compress_type = zipfile.ZIP_DEFLATED
                # very poorly documented:
                info.external_attr = 0o644 << 16
                info.comment = str(tooltip).encode("utf8")
                if isinstance(s, bytes):
                    rm: str = ""
                    try:
                        try:
                            import tempfile
                            temp = tempfile.NamedTemporaryFile(prefix="xpra.", suffix=".%s" % dtype, delete=False)
                            rm = temp.name
                            with temp:
                                temp.write(s)
                                temp.flush()
                        except OSError as e:
                            log.error("Error: cannot create mmap file:")
                            log.estr(e)
                        else:
                            zf.write(temp.name, cfile, zipfile.ZIP_STORED if dtype == "png" else zipfile.ZIP_DEFLATED)
                    finally:
                        if rm:
                            os.unlink(rm)
                else:
                    zf.writestr(info, str(s))
        except OSError as e:
            log("do_save(%s) failed to save zip file", filename, exc_info=True)
            dialog = Gtk.MessageDialog(self.window, 0, Gtk.MessageType.WARNING,
                                       Gtk.ButtonsType.CLOSE, "Failed to save ZIP file")
            dialog.format_secondary_text("%s" % e)

            def close(*_args) -> None:
                dialog.close()

            dialog.connect("response", close)
            dialog.show_all()
        finally:
            if zf:
                zf.close()


def main(argv=()) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.platform import program_context
    from xpra.platform.gui import init, set_default_icon
    from xpra.gtk.util import init_display_source
    with program_context("Xpra-Bug-Report", "Xpra Bug Report"):
        from xpra.log import enable_color
        enable_color()
        init_display_source()
        set_default_icon("bugs.png")
        init()

        from xpra.log import enable_debug_for
        if "-v" in argv:
            enable_debug_for("util")

        from xpra.gtk.signals import register_os_signals
        app = BugReport()
        app.close = app.quit
        app.init(True)
        register_os_signals(app.quit, "Bug Report")
        try:
            from xpra.platform.gui import ready as gui_ready
            gui_ready()
            app.show()
            app.run()
        except KeyboardInterrupt:
            pass
    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv))
